# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log
import netaddr

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class ListNameserversFixture(tobiko.SharedFixture):

    ssh_client: typing.Optional[ssh.SSHClientFixture] = None
    filenames: typing.Optional[typing.Sequence[str]] = None

    nameservers: typing.Optional[tobiko.Selection[netaddr.IPAddress]] = None

    def __init__(self,
                 ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
                 filenames: typing.Optional[typing.Iterable[str]] = None,
                 **execute_params):
        super(ListNameserversFixture, self).__init__()
        if ssh_client is not None:
            self.ssh_client = ssh_client
        if filenames is not None:
            self.filenames = list(filenames)
        self.execute_params = execute_params

    def setup_fixture(self):
        self.nameservers = list_nameservers(ssh_client=self.ssh_client,
                                            filenames=self.filenames,
                                            **self.execute_params)


def list_nameservers(ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
                     filenames: typing.Optional[typing.Iterable[str]] = None,
                     ip_version: typing.Optional[int] = None,
                     **execute_params) -> \
        tobiko.Selection[netaddr.IPAddress]:
    if filenames is None:
        filenames = ['/etc/resolv.conf']

    nameservers: tobiko.Selection[netaddr.IPAddress] = tobiko.Selection()
    for filename in filenames:
        nameservers.extend(parse_resolv_conf_file(ssh_client=ssh_client,
                                                  filename=filename,
                                                  **execute_params))
    if ip_version:
        nameservers = nameservers.with_attributes(version=ip_version)
    return nameservers


def parse_resolv_conf_file(
        filename: str,
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        **execute_params) -> \
        typing.Generator[netaddr.IPAddress, None, None]:
    lines: typing.List[str] = \
        sh.execute(f"cat '{filename}'",
                   ssh_client=ssh_client,
                   **execute_params).stdout.splitlines()
    for line in lines:
        for comment_sep in [';', '#']:
            if comment_sep in line:
                # Filter out comments
                line, _ = line.split(comment_sep, 1)
        # Filter out heading and trailing white spaces
        line = line.strip().lower()
        if not line:
            # Skip empty lines
            continue
        fields = line.split()
        if fields[0] == 'nameserver':
            for nameserver in fields[1:]:
                try:
                    yield netaddr.IPAddress(nameserver)
                except netaddr.AddrFormatError:
                    LOG.exception(f"Invalid nameserver address: {nameserver}")



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

import tobiko
from tobiko.shell import sh


LOG = log.getLogger(__name__)


class RebootHostTest(testtools.TestCase):

    def test_list_processes(self):
        processes = sh.list_processes()
        self._check_processes(processes,
                              is_kernel=False)

    def test_list_kernel_processes(self):
        processes = sh.list_kernel_processes()
        self._check_processes(processes=processes, is_kernel=True)

    def test_list_all_processes(self):
        processes = sh.list_all_processes()
        self._check_processes(processes=processes, is_kernel=None)

    def test_list_processes_with_pid(self):
        processes = sh.list_processes()
        processes_with_pid = sh.list_processes(pid=processes[0].pid)
        self.assertEqual(processes[:1], processes_with_pid)

    def test_list_processes_with_command(self):
        processes = sh.list_processes(command='systemd')
        for process in processes:
            self.assertTrue(process.command.startswith('systemd'), process)

    def test_list_processes_with_exact_command(self):
        processes = sh.list_processes(command='^systemd$')
        self.assertEqual(processes.with_attributes(command='systemd'),
                         processes)

    def _check_processes(self, processes, is_kernel):
        self.assertIsInstance(processes, tobiko.Selection)
        for process in processes:
            self.assertGreater(process.pid, 0)
            self.assertIs(
                (process.command.startswith('[') and
                 process.command.endswith(']')),
                process.is_kernel)
            if is_kernel is not None:
                self.assertIs(bool(is_kernel), process.is_kernel)

    def test_wait_for_processes(self):
        # assume the PID of the first execution of PS process is not more there
        # at the second execution
        process = sh.list_processes(command='ps')[-1]
        sh.wait_for_processes(pid=process.pid, command='ps', timeout=30.)

    def test_wait_for_processes_timeout(self):
        # assume there are always to be running processes on host
        ex = self.assertRaises(sh.PsWaitTimeout, sh.wait_for_processes,
                               timeout=3.)
        self.assertEqual(3., ex.timeout)
        self.assertEqual(sh.get_hostname(), ex.hostname)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc

from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)


class OpenstackClientFixture(tobiko.SharedFixture):

    client = None
    session = None

    def __init__(self, session=None, client=None):
        super(OpenstackClientFixture, self).__init__()
        if session:
            self.session = session
        if client:
            self.client = client

    def setup_fixture(self):
        self.setup_client()

    def setup_client(self):
        client = self.client
        if not client:
            self.session = session = self.get_session()
            self.client = client = self.init_client(session=session)
        return client

    def get_session(self):
        from tobiko.openstack import keystone
        return keystone.keystone_session(self.session)

    @abc.abstractmethod
    def init_client(self, session):
        raise NotImplementedError


class OpenstackClientManager(object):

    def __init__(self):
        self.clients = {}

    def get_client(self, session=None, shared=True, init_client=None):
        if shared:
            if session and tobiko.is_fixture(session):
                key = tobiko.get_fixture_name(session)
            else:
                key = session

            client = self.clients.get(key)
            if client:
                return client

        init_client = init_client or self.create_client
        assert callable(init_client)
        LOG.debug('Initialize OpenStack client: %r(session=%r)',
                  init_client, session)
        client = init_client(session=session)

        if shared:
            self.clients[key] = client
        return client

    def create_client(self, session):
        raise NotImplementedError



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.glance import _client
from tobiko.openstack.glance import _image
from tobiko.openstack.glance import _io
from tobiko.openstack.glance import _lzma


glance_client = _client.glance_client
get_glance_client = _client.get_glance_client
GlanceClientFixture = _client.GlanceClientFixture
create_image = _client.create_image
get_image = _client.get_image
find_image = _client.find_image
list_images = _client.list_images
delete_image = _client.delete_image

FileGlanceImageFixture = _image.FileGlanceImageFixture
GlanceImageFixture = _image.GlanceImageFixture
HasImageMixin = _image.HasImageMixin
URLGlanceImageFixture = _image.URLGlanceImageFixture

open_image_file = _io.open_image_file

has_lzma = _lzma.has_lzma



from __future__ import absolute_import

from oslo_log import log
import pandas
import six

import tobiko
from tobiko.tripleo import overcloud
from tobiko.shell import sh


LOG = log.getLogger(__name__)


class OvercloudServiceException(tobiko.TobikoException):
    message = "not all overcloud nodes services are in active state"


def get_overcloud_node_services_table(hostname):
    """
    get services table from overcloud node

       returns :
auditd.service|loaded|active|running|SecurityAuditingService
auth-rpcgss-module.service|loaded|inactivedead|KernelModulesupportingRPCSEC_GSS
blk-availability.service|loaded|active|exited|Availabilityofblockdevices
brandbot.service|loaded|inactivedead|FlexibleBrandingService
certmonger.service|loaded|active|running|CertificatemonitoringandPKIenrollment
cinder-lvm-losetup.service|loaded|inactivedead|CinderLVMlosetup
cloud-config.service|loaded|active|exited|Applythesettingsspecifiedincloud-con
cloud-final.service|loaded|active|exited|Executeclouduser/finalscripts
cloud-init-local.service|loaded|active|exited|Initialcloud-initjob(pre-network)
cloud-init.service|loaded|active|exited|Initialcloud-initjob(metadataservicecr)

    :return: dataframe of overcloud node services
    """

    ssh_client = overcloud.overcloud_ssh_client(hostname)
    output = sh.execute(
        "systemctl -a --no-pager --plain --no-legend|grep -v not-found|"
        "sed \'s/\\s\\s/|/g\'|sed \'s/||*/DELIM/g\'|sed \'s@ @@g\'|"
        "sed \'s/DELIM$//g\'",
        ssh_client=ssh_client).stdout
    stream = six.StringIO(output)
    table = pandas.read_csv(stream, sep='DELIM', header=None, skiprows=0)
    table.replace(to_replace=' ', value="", regex=True, inplace=True)
    table.columns = ['UNIT', 'loaded_state', 'active_state',
                     'low_level_state', 'UNIT_DESCRIPTION']
    table['overcloud_node'] = hostname

    LOG.debug("Got overcloud nodes services status :\n%s", table)
    return table


def get_overcloud_nodes_running_service(service):
    """
    Check what nodes are running the specified service or unit
    process: exact str of a process name as seen in systemctl -a
    :return: list of overcloud nodes
    """
    oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_services_table)
    oc_nodes_running_service = oc_procs_df.query('UNIT=="{}"'.format(service))[
                                                'overcloud_node'].unique()
    return oc_nodes_running_service


def check_if_process_running_on_overcloud(process):
    """
    Check what nodes are running the specifies
    process: exact str of a process name as seen in ps -axw -o "%c"
    :return: list of overcloud nodes
    """
    oc_services_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_services_table)
    if not oc_services_df.query('UNIT=="{}"'.format(process)).empty:
        return True
    else:
        return False


class OvercloudServicesStatus(object):
    """
    class to handle services checks,
    checks that all of these are running in the overcloud:
    'corosync.service','iptables.service','network.service','ntpd.service',
    'pacemaker.service','rpcbind.service','sshd.service'

    """
    def __init__(self):
        self.services_to_check = ['corosync.service', 'iptables.service',
                                  'network.service', 'ntpd.service',
                                  'pacemaker.service', 'rpcbind.service',
                                  'sshd.service']

        self.oc_services_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_services_table)

    @property
    def basic_overcloud_services_running(self):
        """
        Checks that the oc_services dataframe has all of the list services
        running
        :return: Bool
        """
        for service_name in self.services_to_check:
            if not self.oc_services_df.query('UNIT=="{}"'.format(
                    service_name)).empty:
                LOG.info("overcloud processes status checks: process {} is  "
                         "in running state".format(service_name))
                continue
            else:
                LOG.info("Failure : overcloud processes status checks: "
                         "process {} is not running ".format(service_name))
                raise OvercloudServiceException()
        return True



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
import yaml

from tobiko.openstack import nova


class TestUserData(testtools.TestCase):

    def test_user_data(self):
        self.assertEqual('', nova.user_data())

    def test_user_data_with_packages(self):
        user_data = nova.user_data({'packages': [1, 2]}, packages=[2, 3])
        self.assert_equal_cloud_config({"packages": [1, 2, 3]},
                                       user_data)

    def test_user_data_with_runcmd(self):
        user_data = nova.user_data({'runcmd': [["echo", 1]]},
                                   runcmd=['echo 2'])
        self.assert_equal_cloud_config({'runcmd': [['echo', '1'],
                                                   ['echo', '2']]},
                                       user_data)

    def assert_equal_cloud_config(self, expected, actual):
        self.assertTrue(actual.startswith('#cloud-config'))
        self.assertEqual(expected, yaml.load(actual))


class TestCloudConfig(testtools.TestCase):

    def test_cloud_config(self):
        self.assertEqual({}, nova.cloud_config())

    def test_cloud_config_with_packages(self):
        cloud_config = nova.cloud_config({'packages': [1, 2]}, packages=[2, 3])
        self.assertEqual({"packages": [1, 2, 3]}, cloud_config)

    def test_cloud_config_with_runcmd(self):
        cloud_config = nova.cloud_config({'runcmd': [["echo", 1]]},
                                         runcmd=['echo 2'])
        self.assertEqual({'runcmd': [['echo', '1'],
                                     ['echo', '2']]},
                         cloud_config)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools

import tobiko
from tobiko.openstack import glance
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.openstack import tests


class MyServerStack(tests.TestServerCreationStack):
    pass


@keystone.skip_unless_has_keystone_credentials()
class ServerCreationTest(testtools.TestCase):

    def test_server_creation(self):
        all_servers_ids = {server.id for server in nova.list_servers()}

        stack = tests.test_server_creation()

        self.assertIsInstance(stack, tests.TestServerCreationStack)
        class_name = tobiko.get_fixture_name(tests.TestServerCreationStack)
        pid = os.getpid()
        self.assertEqual(f"{class_name}-{pid}-0", stack.stack_name)
        self.assertNotIn(stack.server_id, all_servers_ids)
        self.assertEqual('ACTIVE', nova.get_server(stack.server_id).status)

    def test_server_creation_with_stack(self):
        stack = tests.test_server_creation(stack=MyServerStack)
        self.assertIsInstance(stack, tests.TestServerCreationStack)

    def test_evacuable_server_creation(self):
        stack = tests.test_evacuable_server_creation()
        self.assertIsInstance(stack, tests.TestEvacuableServerCreationStack)
        self.assertIn('evacuable', glance.get_image(stack.image).tags)

    def test_server_creation_and_shutoff(self):
        stack = tests.test_server_creation_and_shutoff()
        self.assertIsInstance(stack, tests.TestServerCreationStack)
        self.assertEqual('SHUTOFF', nova.get_server(stack.server_id).status)

    def test_servers_creation(self):
        all_servers_ids = {server.id for server in nova.list_servers()}
        stacks = tests.test_servers_creation()
        self.assertEqual(2, len(stacks))
        pid = os.getpid()
        for i, stack in enumerate(stacks):
            self.assertIsInstance(stack, tests.TestServerCreationStack)
            class_name = tobiko.get_fixture_name(tests.TestServerCreationStack)
            self.assertEqual(f"{class_name}-{pid}-{i}", stack.stack_name)
            self.assertNotIn(stack.server_id, all_servers_ids)
            self.assertEqual('ACTIVE', nova.get_server(stack.server_id).status)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg

GROUP_NAME = 'shell'
OPTIONS = [
    cfg.StrOpt('command',
               default='/bin/sh -c',
               help="Default shell command used for executing "
                    "local commands"),
    cfg.StrOpt('sudo',
               default='sudo',
               help="Default sudo command used for executing "
                    "commands as superuser or another user")
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup('shell'), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

from tobiko import config
from tobiko.shell import sh
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko import tripleo


CONF = config.CONF


@tripleo.skip_if_missing_undercloud
class UndercloudSshConnectionTest(testtools.TestCase):

    def setUp(self):
        super(UndercloudSshConnectionTest, self).setUp()
        self.ssh_client = tripleo.undercloud_ssh_client()

    def test_connect_to_undercloud(self):
        self.ssh_client.connect()

    def test_fetch_undercloud_credentials(self):
        env = tripleo.load_undercloud_rcfile()
        self.assertTrue(env['OS_AUTH_URL'])
        self.assertTrue(env.get('OS_USERNAME') or env.get('OS_USER_ID'))
        self.assertTrue(env['OS_PASSWORD'])
        self.assertTrue(env.get('OS_TENANT_NAME') or
                        env.get('OS_PROJECT_NAME') or
                        env.get('OS_TENANT_ID') or
                        env.get('OS_PROJECT_ID'))

    def test_execute_command(self):
        result = sh.execute('hostname', ssh_client=self.ssh_client)
        self.assertTrue(result.stdout.startswith('undercloud-0'))


@tripleo.skip_if_missing_undercloud
class UndercloudKeystoneClientTest(testtools.TestCase):

    def test_undercloud_keystone_credentials(self):
        credentials = tripleo.undercloud_keystone_credentials()
        self.assertIsInstance(credentials, keystone.KeystoneCredentials)
        credentials.validate()

    def test_undercloud_keystone_session(self):
        session = tripleo.undercloud_keystone_session()
        client = nova.get_nova_client(session=session)
        overcloud_nodes = nova.list_servers(client=client)
        self.assertTrue(overcloud_nodes)

    def test_undercloud_keystone_client(self):
        client = tripleo.undercloud_keystone_client()
        services = keystone.list_services(client=client)
        self.assertTrue(services)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack.stacks import _centos
from tobiko.openstack.stacks import _cirros
from tobiko.openstack.stacks import _neutron
from tobiko.openstack.stacks import _nova
from tobiko.openstack.stacks import _ubuntu


@neutron.skip_if_missing_networking_extensions('l3-ha')
class L3haNetworkStackFixture(_neutron.NetworkStackFixture):
    ha = True


@neutron.skip_if_missing_networking_extensions('l3-ha')
class L3haServerStackFixture(_cirros.CirrosServerStackFixture):
    #: Heat stack for creating internal network with L3HA enabled
    network_stack = tobiko.required_setup_fixture(
        L3haNetworkStackFixture)


class L3haPeerServerStackFixture(
        L3haServerStackFixture, _nova.PeerServerStackFixture):
    peer_stack = tobiko.required_setup_fixture(L3haServerStackFixture)


class L3haSameHostServerStackFixture(
        L3haPeerServerStackFixture, _nova.SameHostServerStackFixture):
    pass


class L3haDifferentHostServerStackFixture(
        L3haPeerServerStackFixture, _nova.DifferentHostServerStackFixture):
    pass


class L3haCentosServerStackFixture(_centos.CentosServerStackFixture,
                                   L3haServerStackFixture):
    pass


class L3haUbuntuServerStackFixture(_ubuntu.UbuntuServerStackFixture,
                                   L3haServerStackFixture):
    pass



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import typing  # noqa

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


class NoMatchingLinesFound(tobiko.TobikoException):
    message = ("No matching lines found in files (pattern='{pattern}',"
               " files={files}, login={login})")


def grep_files(pattern: str,
               files: typing.List[str],
               command: sh.ShellCommandType = 'zgrep -Eh',
               ssh_client: ssh.SSHClientFixture = None,
               blank_lines=False,
               **execute_params) -> typing.List[str]:
    if not pattern:
        raise ValueError("Pattern string can't be empty")
    if not files:
        raise ValueError("File list can't be empty")
    command_line = sh.shell_command(command) + ['-e', pattern] + files
    try:
        result = sh.execute(command_line,
                            ssh_client=ssh_client,
                            **execute_params)
    except sh.ShellCommandFailed as ex:
        if ex.exit_status > 1:
            # Some unknown problem occurred
            raise
    else:
        output_lines: typing.List[str] = [
            line
            for line in result.stdout.splitlines()
            if blank_lines or line.strip()]
        if output_lines:
            return output_lines
    raise NoMatchingLinesFound(pattern=pattern,
                               files=files,
                               login=ssh_client and ssh_client.login or None)


def grep_lines(pattern: str,
               command: sh.ShellCommandType,
               ssh_client: ssh.SSHClientFixture = None,
               **execute_params) -> typing.List[str]:
    if not pattern:
        raise ValueError("Pattern string can't be empty")
    command_line = sh.shell_command(command)
    try:
        result = sh.execute(command_line,
                            ssh_client=ssh_client,
                            **execute_params)
    except sh.ShellCommandFailed as ex:
        if ex.exit_status > 1:
            # Some unknown problem occurred
            raise
    else:
        output_lines: typing.List[str] = []
        r = re.compile(pattern)
        output_lines = [line for line in result.stdout.splitlines()
                        if r.search(line)]
        if output_lines:
            return output_lines
    raise NoMatchingLinesFound(pattern=pattern,
                               files=command,
                               login=ssh_client and ssh_client.login or None)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import shlex
import typing  # noqa


class ShellCommand(tuple):

    def __repr__(self):
        return "ShellCommand([{!s}])".format(', '.join(self))

    def __str__(self):
        return list_to_command_line(self)

    def __add__(self, other):
        other = shell_command(other)
        return shell_command(tuple(self) + other)


ShellCommandType = typing.Union[ShellCommand, str, typing.Iterable]


def shell_command(command: ShellCommandType) -> ShellCommand:
    if isinstance(command, ShellCommand):
        return command
    elif isinstance(command, str):
        return ShellCommand(shlex.split(command))
    else:
        return ShellCommand(str(a) for a in command)


def list_to_command_line(seq):
    result = []
    for arg in seq:
        bs_buf = []

        # Add a space to separate this argument from the others
        if result:
            result.append(' ')

        needquote = (" " in arg) or ("\t" in arg) or not arg
        if needquote:
            result.append("'")

        for c in arg:
            if c == '\\':
                # Don't know if we need to double yet.
                bs_buf.append(c)
            elif c == '"':
                # Double backslashes.
                result.append('\\' * len(bs_buf)*2)
                bs_buf = []
                result.append('\\"')
            else:
                # Normal char
                if bs_buf:
                    result.extend(bs_buf)
                    bs_buf = []
                result.append(c)

        # Add remaining backslashes, if any.
        if bs_buf:
            result.extend(bs_buf)

        if needquote:
            result.extend(bs_buf)
            result.append("'")

    return ''.join(result)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.stacks import _centos
from tobiko.openstack.stacks import _cirros
from tobiko.openstack.stacks import _redhat
from tobiko.openstack.stacks import _l3ha
from tobiko.openstack.stacks import _neutron
from tobiko.openstack.stacks import _nova
from tobiko.openstack.stacks import _octavia
from tobiko.openstack.stacks import _ubuntu

CentosFlavorStackFixture = _centos.CentosFlavorStackFixture
CentosImageFixture = _centos.CentosImageFixture
CentosServerStackFixture = _centos.CentosServerStackFixture

CirrosFlavorStackFixture = _cirros.CirrosFlavorStackFixture
CirrosImageFixture = _cirros.CirrosImageFixture
CirrosServerStackFixture = _cirros.CirrosServerStackFixture
CirrosExternalServerStackFixture = _cirros.CirrosExternalServerStackFixture
CirrosPeerServerStackFixture = _cirros.CirrosPeerServerStackFixture
CirrosDifferentHostServerStackFixture = (
    _cirros.CirrosDifferentHostServerStackFixture)
CirrosSameHostServerStackFixture = _cirros.CirrosSameHostServerStackFixture
RebootCirrosServerOperation = _cirros.RebootCirrosServerOperation
EvacuableCirrosImageFixture = _cirros.EvacuableCirrosImageFixture
EvacuableServerStackFixture = _cirros.EvacuableServerStackFixture

RedHatFlavorStackFixture = _redhat.RedHatFlavorStackFixture
RhelImageFixture = _redhat.RhelImageFixture
RedHatServerStackFixture = _redhat.RedHatServerStackFixture

L3haNetworkStackFixture = _l3ha.L3haNetworkStackFixture
L3haServerStackFixture = _l3ha.L3haServerStackFixture
L3haUbuntuServerStackFixture = _l3ha.L3haUbuntuServerStackFixture
L3haCentosServerStackFixture = _l3ha.L3haCentosServerStackFixture
L3haPeerServerStackFixture = _l3ha.L3haPeerServerStackFixture
L3haDifferentHostServerStackFixture = _l3ha.L3haDifferentHostServerStackFixture
L3haSameHostServerStackFixture = _l3ha.L3haSameHostServerStackFixture

NetworkStackFixture = _neutron.NetworkStackFixture
NetworkWithNetMtuWriteStackFixture = (
    _neutron.NetworkWithNetMtuWriteStackFixture)
SecurityGroupsFixture = _neutron.SecurityGroupsFixture

get_external_network = _neutron.get_external_network
has_external_network = _neutron.has_external_network
skip_unless_has_external_network = _neutron.skip_unless_has_external_network
get_floating_network = _neutron.get_floating_network
has_floating_network = _neutron.has_floating_network
skip_unless_has_floating_network = _neutron.skip_unless_has_floating_network

ServerStackFixture = _nova.ServerStackFixture
KeyPairStackFixture = _nova.KeyPairStackFixture
FlavorStackFixture = _nova.FlavorStackFixture
ServerGroupStackFixture = _nova.ServerGroupStackFixture
AffinityServerGroupStackFixture = _nova.AffinityServerGroupStackFixture
AntiAffinityServerGroupStackFixture = _nova.AntiAffinityServerGroupStackFixture

UbuntuFlavorStackFixture = _ubuntu.UbuntuFlavorStackFixture
UbuntuImageFixture = _ubuntu.UbuntuImageFixture
UbuntuServerStackFixture = _ubuntu.UbuntuServerStackFixture

OctaviaLoadbalancerStackFixture = _octavia.OctaviaLoadbalancerStackFixture
OctaviaListenerStackFixture = _octavia.OctaviaListenerStackFixture
OctaviaMemberServerStackFixture = _octavia.OctaviaMemberServerStackFixture
OctaviaServerStackFixture = _octavia.OctaviaServerStackFixture
OctaviaClientServerStackFixture = _octavia.OctaviaClientServerStackFixture



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import os

from oslo_log import log
import yaml

import tobiko
from tobiko.openstack.keystone import _credentials


LOG = log.getLogger(__name__)

YAML_SUFFIXES = ('.yaml', '.yml')
JSON_SUFFIXES = ('.json',)
CLOUDS_FILE_SUFFIXES = JSON_SUFFIXES + YAML_SUFFIXES


class CloudsFileNotFoundError(tobiko.TobikoException):
    message = "No such clouds file(s): {clouds_files!s}"


class DefaultCloudsFileConfig(tobiko.SharedFixture):

    cloud_name = None
    clouds_file_dirs = None
    clouds_file_names = None
    clouds_files = None

    def setup_fixture(self):
        keystone_conf = tobiko.tobiko_config().keystone
        self.cloud_name = keystone_conf.cloud_name
        self.clouds_file_dirs = keystone_conf.clouds_file_dirs
        self.clouds_file_names = keystone_conf.clouds_file_names
        self.clouds_files = self.list_cloud_files()

    def list_cloud_files(self):
        cloud_files = []
        for directory in self.clouds_file_dirs:
            directory = tobiko.tobiko_config_path(directory)
            if os.path.isdir(directory):
                for file_name in self.clouds_file_names:
                    file_name = os.path.join(directory, file_name)
                    if os.path.isfile(file_name):
                        cloud_files.append(file_name)
        return cloud_files


class CloudsFileKeystoneCredentialsFixture(
        _credentials.KeystoneCredentialsFixture):

    cloud_name = None
    clouds_content = None
    clouds_file = None

    config = tobiko.required_setup_fixture(DefaultCloudsFileConfig)

    def __init__(self, credentials=None, cloud_name=None,
                 clouds_content=None, clouds_file=None, clouds_files=None):
        super(CloudsFileKeystoneCredentialsFixture, self).__init__(
            credentials=credentials)

        config = self.config
        if cloud_name is None:
            cloud_name = config.cloud_name
        self.cloud_name = cloud_name

        if clouds_content is not None:
            self.clouds_content = dict(clouds_content)

        if clouds_file is not None:
            self.clouds_file = clouds_file

        if clouds_files is None:
            clouds_files = config.clouds_files
        self.clouds_files = list(clouds_files)

    def get_credentials(self):
        cloud_name = self._get_cloud_name()
        if cloud_name is None:
            return None

        clouds_content = self._get_clouds_content()
        clouds_section = clouds_content.get("clouds")
        if clouds_section is None:
            message = ("'clouds' section not found in clouds file "
                       "{!r}").format(self.clouds_file)
            raise ValueError(message)

        clouds_config = clouds_section.get(cloud_name)
        if clouds_config is None:
            message = ("No such cloud with name {!r} in file "
                       "{!r}").format(cloud_name, self.clouds_file)
            raise ValueError(message)

        auth = clouds_config.get("auth")
        if auth is None:
            message = ("No such 'auth' section in cloud file {!r} for cloud "
                       "name {!r}").format(self.clouds_file, self.cloud_name)
            raise ValueError(message)

        auth_url = auth.get("auth_url")
        if not auth_url:
            message = ("No such 'auth_url' in file {!r} for cloud name "
                       "{!r}").format(self.clouds_file, self.cloud_name)
            raise ValueError(message)

        username = auth.get('username') or auth.get('user_id')
        password = auth.get('password')
        cacert = clouds_config.get('cacert')
        project_name = (auth.get('project_name') or
                        auth.get('tenant_namer') or
                        auth.get('project_id') or
                        auth.get_env('tenant_id'))

        api_version = (int(clouds_config.get("identity_api_version", 0)) or
                       _credentials.api_version_from_url(auth_url))
        if api_version == 2:
            return _credentials.keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name)

        else:
            domain_name = (auth.get("domain_name") or
                           auth.get("domain_id"))
            user_domain_name = (auth.get("user_domain_name") or
                                auth.get("user_domain_id"))
            project_domain_name = auth.get("project_domain_name")
            project_domain_id = auth.get("project_domain_id")
            trust_id = auth.get("trust_id")
            return _credentials.keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name,
                domain_name=domain_name,
                user_domain_name=user_domain_name,
                project_domain_name=project_domain_name,
                project_domain_id=project_domain_id,
                cacert=cacert,
                trust_id=trust_id)

    def _get_cloud_name(self):
        cloud_name = self.cloud_name
        if cloud_name is None:
            cloud_name = os.environ.get("OS_CLOUD")
            if cloud_name:
                LOG.debug("Got cloud name from 'OS_CLOUD' environment "
                          "variable: %r", cloud_name)
                self.cloud_name = cloud_name
            else:
                LOG.debug("Undefined environment variable: 'OS_CLOUD'")
        return cloud_name or None

    def _get_clouds_content(self):
        clouds_content = self.clouds_content
        if clouds_content is None:
            clouds_file = self._get_clouds_file()
            with open(clouds_file, 'r') as f:
                _, suffix = os.path.splitext(clouds_file)
                if suffix in JSON_SUFFIXES:
                    LOG.debug('Load JSON clouds file: %r', clouds_file)
                    clouds_content = json.load(f)
                else:
                    LOG.debug('Load YAML clouds file: %r', clouds_file)
                    clouds_content = yaml.safe_load(f)
            LOG.debug('Clouds file content loaded from %r:\n%s',
                      clouds_file, json.dumps(clouds_content,
                                              indent=4,
                                              sort_keys=True))
            self.clouds_content = clouds_content

        if not clouds_content:
            message = "Invalid clouds file content: {!r}".format(
                clouds_content)
            raise ValueError(message)
        return clouds_content

    def _get_clouds_file(self):
        clouds_file = self.clouds_file
        if clouds_file:
            clouds_files = [self.clouds_file]
        else:
            clouds_files = list(self.clouds_files)

        for filename in clouds_files:
            if os.path.exists(filename):
                LOG.debug('Found clouds file at %r', filename)
                self.clouds_file = clouds_file = filename
                break
        else:
            raise CloudsFileNotFoundError(clouds_files=', '.join(clouds_files))
        return clouds_file


_credentials.DEFAULT_KEYSTONE_CREDENTIALS_FIXTURES.insert(
    0, CloudsFileKeystoneCredentialsFixture)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import sys

from oslo_log import log
import six

import tobiko
from tobiko.cmd import base

LOG = log.getLogger(__name__)


class FixtureUtil(base.TobikoCMD):

    def get_parser(self):
        parser = super(FixtureUtil, self).get_parser()

        subparsers = parser.add_subparsers(
            title='subcommands', description='fixture operation')

        subcommand_parser = subparsers.add_parser(
            'help', help=('show this help message'))
        subcommand_parser.set_defaults(subcommand='help')

        for subcommand_name in ['cleanup', 'list', 'setup']:
            subcommand_parser = subparsers.add_parser(
                subcommand_name, help=(subcommand_name + ' fixtures'))
            subcommand_parser.set_defaults(subcommand=subcommand_name)
            subcommand_parser.add_argument(
                '--config', '-c',
                default='.stestr.conf',
                help=("Set a stestr config file to use with this command. "
                      "If one isn't specified then .stestr.conf in the "
                      "directory that a command is running from is used"))
            subcommand_parser.add_argument(
                '--repo-type', '-r',
                choices=['file'],
                default='file',
                help=("Select the repo backend to use"))
            subcommand_parser.add_argument(
                '--repo-url', '-u',
                help=("Set the repo url to use. An acceptable value for "
                      "this depends on the repository type used."))
            subcommand_parser.add_argument(
                '--test-path', '-t',
                help=("Set the test path to use for unittest discovery. If "
                      "both this and the corresponding config file option "
                      "are set, this value will be used."))
            subcommand_parser.add_argument(
                '--top-dir',
                help=("Set the top dir to use for unittest discovery. If "
                      "both this and the corresponding config file option "
                      "are set, this value will be used."))
            subcommand_parser.add_argument(
                '--group-regex', '--group_regex', '-g',
                help=("Set a group regex to use for grouping tests together "
                      "in the stestr scheduler. If both this and the "
                      "corresponding config file option are set this value "
                      "will be used."))
            subcommand_parser.add_argument(
                '--blacklist-file', '-b',
                help=("Path to a blacklist file, this file contains a "
                      "separate regex exclude on each newline."))
            subcommand_parser.add_argument(
                '--whitelist-file', '-w',
                help=("Path to a whitelist file, this file contains a "
                      "separate regex on each newline."))
            subcommand_parser.add_argument(
                '--black-regex', '-B',
                help=("Test rejection regex. If a test cases name matches "
                      "on re.search() operation , it will be removed from "
                      "the final test list. Effectively the black-regexp is "
                      "added to black regexp list, but you do need to edit "
                      "a file. The black filtering happens after the "
                      "initial white selection, which by default is "
                      "everything."))
            subcommand_parser.add_argument(
                'filters',
                nargs='*',
                help=("A list of string regex filters to initially apply "
                      "on the test list. Tests that match any of the "
                      "regexes will be used. (assuming any other filtering "
                      "specified also uses it)."))
        return parser

    def execute(self):
        action = self.args.subcommand or 'list'
        if action == 'help':
            return self.show_help()
        if action == 'list':
            return self.list_fixtures()
        elif action == 'setup':
            return self.setup_fixtures()
        elif action == 'cleanup':
            return self.cleanup_fixtures()

    def show_help(self, stream=None):
        stream = stream or sys.stdout
        self.parser.print_help(stream)

    def discover_testcases(self):
        args = self.args
        return tobiko.discover_test_cases(
            config=args.config,
            repo_type=args.repo_type,
            repo_url=args.repo_url,
            test_path=args.test_path,
            top_dir=args.top_dir,
            group_regex=args.group_regex,
            blacklist_file=args.blacklist_file,
            whitelist_file=args.whitelist_file,
            black_regex=args.black_regex,
            filters=args.filters)

    def list_fixtures(self, stream=None):
        stream = stream or sys.stdout
        test_cases = self.discover_testcases()
        fixtures_names = tobiko.list_required_fixtures(test_cases)
        if fixtures_names:
            output = '\n'.join(fixtures_names) + '\n'
            if six.PY2:
                output = output.decode()
            stream.write(output)

    def setup_fixtures(self, stream=None):
        stream = stream or sys.stdout
        test_cases = self.discover_testcases()
        success = True
        for fixture_name in tobiko.list_required_fixtures(test_cases):
            output = fixture_name + '\n'
            if six.PY2:
                output = output.decode()
            stream.write(output)
            try:
                tobiko.setup_fixture(fixture_name)
            except Exception:
                LOG.exception('Fixture %r setup failed', fixture_name)
                success = False
        if not success:
            sys.exit(1)

    def cleanup_fixtures(self, stream=None):
        stream = stream or sys.stdout
        test_cases = self.discover_testcases()
        success = True
        for fixture_name in tobiko.list_required_fixtures(test_cases):
            output = fixture_name + '\n'
            if six.PY2:
                output = output.decode()
            stream.write(output)
            try:
                tobiko.cleanup_fixture(fixture_name)
            except Exception:
                LOG.exception('Fixture %r cleanup failed', fixture_name)
                success = False
        if not success:
            sys.exit(1)


def main():
    """Create CLI main entry."""
    fixture_util = FixtureUtil()
    fixture_util.set_stream_handler_logging_level()
    fixture_util.execute()


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.docker import _exception
from tobiko.shell import sh


def discover_docker_urls(**execute_params):
    result = sh.execute('ps aux | grep dockerd', stdin=False, stdout=True,
                        stderr=True, expect_exit_status=None, **execute_params)
    if result.exit_status or not result.stdout:
        raise _exception.DockerUrlNotFoundError(details=result.stderr)

    urls = []
    for line in result.stdout.splitlines():
        fields = line.strip().split()
        if fields:
            offset = 0
            while True:
                try:
                    offset = fields.index('-H', offset)
                    url = fields[offset + 1]
                except (ValueError, IndexError):
                    break
                else:
                    urls.append(url)
                    offset += 2

    if not urls:
        raise _exception.DockerUrlNotFoundError(details='\n' + result.stdout)

    return urls


def is_docker_running(ssh_client=None, **execute_params):
    try:
        discover_docker_urls(ssh_client=ssh_client, **execute_params)
    except _exception.DockerUrlNotFoundError:
        return False
    else:
        return True



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.tests import unit


class TestFail(unit.TobikoUnitTest):

    def test_fail(self):
        self._test_fail('some_reason')

    def test_fail_with_args(self):
        self._test_fail('some {1!r} {0!s}', 'reason', 'strange')

    def test_fail_with_kwargs(self):
        self._test_fail('some {b!r} {a!s}', a='reason', b='strange')

    def _test_fail(self, reason, *args, **kwargs):
        ex = self.assertRaises(
            tobiko.FailureException, tobiko.fail, reason, *args, **kwargs)
        if args or kwargs:
            expected_reason = reason.format(*args, **kwargs)
        else:
            expected_reason = reason
        self.assertEqual(expected_reason, str(ex))



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack.stacks import _nova
from tobiko.shell import sh


CONF = config.CONF

CIRROS_IMAGE_VERSION = '0.5.1'

CIRROS_IMAGE_URL = (
    'http://download.cirros-cloud.net/{version}/'
    'cirros-{version}-x86_64-disk.img').format(version=CIRROS_IMAGE_VERSION)


class CirrosImageFixture(glance.URLGlanceImageFixture):

    image_url = CONF.tobiko.cirros.image_url or CIRROS_IMAGE_URL
    image_name = CONF.tobiko.cirros.image_name
    image_file = CONF.tobiko.cirros.image_file
    container_format = CONF.tobiko.cirros.container_format or "bare"
    disk_format = CONF.tobiko.cirros.disk_format or "qcow2"
    username = CONF.tobiko.cirros.username or 'cirros'
    password = CONF.tobiko.cirros.password or 'gocubsgo'
    connection_timeout = CONF.tobiko.cirros.connection_timeout or 200.


class CirrosFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 128


class CirrosServerStackFixture(_nova.ServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_setup_fixture(CirrosImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_setup_fixture(CirrosFlavorStackFixture)


class CirrosPeerServerStackFixture(CirrosServerStackFixture,
                                   _nova.PeerServerStackFixture):
    #: Peer server used to reach this one
    peer_stack = tobiko.required_setup_fixture(CirrosServerStackFixture)


class CirrosSameHostServerStackFixture(
        CirrosPeerServerStackFixture, _nova.SameHostServerStackFixture):
    pass


class CirrosDifferentHostServerStackFixture(
        CirrosPeerServerStackFixture, _nova.DifferentHostServerStackFixture):
    pass


class RebootCirrosServerOperation(sh.RebootHostOperation):

    stack = tobiko.required_setup_fixture(CirrosServerStackFixture)

    @property
    def ssh_client(self):
        return self.stack.ssh_client


class EvacuableCirrosImageFixture(CirrosImageFixture):

    tags = ['evacuable']


class EvacuableServerStackFixture(CirrosServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_setup_fixture(EvacuableCirrosImageFixture)


class CirrosExternalServerStackFixture(CirrosServerStackFixture,
                                       _nova.ExternalServerStackFixture):
    pass



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from docker import client as docker_client
from docker.models import containers
import testtools

import tobiko
from tobiko import docker
from tobiko.openstack import keystone
from tobiko.openstack import topology


class DockerNodeFixture(tobiko.SharedFixture):

    node = None

    def setup_fixture(self):
        nodes = topology.list_openstack_nodes()
        for node in nodes:
            assert node.ssh_client is not None
            if docker.is_docker_running(ssh_client=node.ssh_client):
                self.node = node
                break

        if self.node is None:
            nodes_text = ' '.join(node.name for node in nodes)
            tobiko.skip_test("Docker server is not running in any of nodes "
                             f"{nodes_text}")


@keystone.skip_unless_has_keystone_credentials()
class DockerClientTest(testtools.TestCase):

    node = tobiko.required_setup_fixture(DockerNodeFixture)

    @property
    def ssh_client(self):
        return self.node.node.ssh_client

    def test_get_docker_client(self):
        client = docker.get_docker_client(ssh_client=self.ssh_client)
        self.assertIsInstance(client, docker.DockerClientFixture)

    def test_connect_docker_client(self):
        client = docker.get_docker_client(ssh_client=self.ssh_client).connect()
        self.assertIsInstance(client, docker_client.DockerClient)
        client.ping()

    def test_list_docker_containers(self):
        for container in docker.list_docker_containers(
                ssh_client=self.ssh_client):
            self.assertIsInstance(container, containers.Container)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import os

from oslo_log import log
import pandas
import six

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import _undercloud


CONF = config.CONF
LOG = log.getLogger(__name__)


def has_overcloud():
    # rewrite this function
    return _undercloud.has_undercloud()


def load_overcloud_rcfile():
    return _undercloud.fetch_os_env(*CONF.tobiko.tripleo.overcloud_rcfile)


skip_if_missing_overcloud = tobiko.skip_unless(
    'TripleO overcloud not configured', has_overcloud)


class OvercloudKeystoneCredentialsFixture(
        keystone.EnvironKeystoneCredentialsFixture):

    def get_environ(self):
        if has_overcloud():
            return load_overcloud_rcfile()
        else:
            return {}


def list_overcloud_nodes(**params):
    session = _undercloud.undercloud_keystone_session()
    client = nova.get_nova_client(session=session)
    return nova.list_servers(client=client, **params)


def find_overcloud_node(**params):
    session = _undercloud.undercloud_keystone_session()
    client = nova.get_nova_client(session=session)
    return nova.find_server(client=client, **params)


def overcloud_ssh_client(hostname, ip_version=None, network_name=None):
    host_config = overcloud_host_config(hostname=hostname,
                                        ip_version=ip_version,
                                        network_name=network_name)
    return ssh.ssh_client(host=hostname, host_config=host_config)


def overcloud_host_config(hostname, ip_version=None, network_name=None):
    host_config = OvercloudHostConfig(host=hostname,
                                      ip_version=ip_version,
                                      network_name=network_name)
    return tobiko.setup_fixture(host_config)


def overcloud_node_ip_address(ip_version=None, network_name=None, server=None,
                              **params):
    server = server or find_overcloud_node(**params)
    ip_version = ip_version or CONF.tobiko.tripleo.overcloud_ip_version
    network_name = network_name or CONF.tobiko.tripleo.overcloud_network_name
    address = nova.find_server_ip_address(server=server,
                                          ip_version=ip_version,
                                          network_name=network_name)
    LOG.debug(f"Got Overcloud node address '{address}' from Undercloud "
              f"(ip_version={ip_version}, network_name={network_name}, "
              f"server={server})")
    return address


class OvercloudSshKeyFileFixture(tobiko.SharedFixture):

    @property
    def key_filename(self):
        return tobiko.tobiko_config_path(
            CONF.tobiko.tripleo.overcloud_ssh_key_filename)

    def setup_fixture(self):
        self.setup_key_file()

    def setup_key_file(self):
        key_filename = self.key_filename
        key_dirname = os.path.dirname(key_filename)
        tobiko.makedirs(key_dirname, mode=0o700)

        ssh_client = _undercloud.undercloud_ssh_client()
        _get_undercloud_file(ssh_client=ssh_client,
                             source='~/.ssh/id_rsa',
                             destination=key_filename,
                             mode=0o600)
        _get_undercloud_file(ssh_client=ssh_client,
                             source='~/.ssh/id_rsa.pub',
                             destination=key_filename + '.pub',
                             mode=0o600)


def _get_undercloud_file(ssh_client, source, destination, mode):
    content = sh.execute(['cat', source],
                         ssh_client=ssh_client).stdout
    with io.open(destination, 'wb') as fd:
        fd.write(content.encode())
    os.chmod(destination, mode)


class OvercloudHostConfig(tobiko.SharedFixture):
    hostname = None
    port = None
    username = None
    key_file = tobiko.required_setup_fixture(OvercloudSshKeyFileFixture)
    ip_version = None
    network_name = None
    key_filename = None

    def __init__(self, host, ip_version=None, network_name=None,
                 **kwargs):
        super(OvercloudHostConfig, self).__init__()
        tobiko.check_valid_type(host, six.string_types)
        self.host = host
        if ip_version:
            self.ip_version = ip_version
        if network_name:
            self.network_name = network_name
        self._connect_parameters = ssh.gather_ssh_connect_parameters(**kwargs)

    def setup_fixture(self):
        self.hostname = str(overcloud_node_ip_address(
            name=self.host, ip_version=self.ip_version,
            network_name=self.network_name))
        self.port = self.port or CONF.tobiko.tripleo.overcloud_ssh_port
        self.username = (self.username or
                         CONF.tobiko.tripleo.overcloud_ssh_username)
        self.key_filename = self.key_filename or self.key_file.key_filename

    @property
    def connect_parameters(self):
        parameters = ssh.ssh_host_config(
            host=str(self.hostname)).connect_parameters
        parameters.update(ssh.gather_ssh_connect_parameters(self))
        parameters.update(self._connect_parameters)
        return parameters


def setup_overcloud_keystone_crederntials():
    keystone.DEFAULT_KEYSTONE_CREDENTIALS_FIXTURES.append(
        OvercloudKeystoneCredentialsFixture)


def get_overcloud_nodes_dataframe(oc_node_df_function):
    """
     :param oc_node_df_function : a function that queries a oc node
     using a cli command and returns a datraframe with an added
     hostname field.

     This function concats oc nodes dataframes into a unified overcloud
     dataframe, seperated by hostname field

    :return: dataframe of all overcloud nodes processes
    """
    oc_nodes_selection = list_overcloud_nodes()
    oc_nodes_names = [node.name for node in oc_nodes_selection]
    oc_nodes_dfs = [oc_node_df_function(node_name) for
                    node_name in oc_nodes_names]
    oc_procs_df = pandas.concat(oc_nodes_dfs, ignore_index=True)
    return oc_procs_df



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import stacks
from tobiko.shell import sh


CONF = config.CONF


class ExecuteTest(testtools.TestCase):

    ssh_client = None
    shell = '/bin/sh -c'

    def test_succeed(self, command='true', stdin=None, stdout=None,
                     stderr=None, expect_exit_status=0, **kwargs):
        process = self.execute(command=command,
                               stdin=stdin,
                               stdout=bool(stdout),
                               stderr=bool(stderr),
                               expect_exit_status=expect_exit_status,
                               **kwargs)
        self.assertEqual(self.expected_command(command), process.command)
        if stdin:
            self.assertEqual(stdin, str(process.stdin))
        else:
            self.assertIsNone(process.stdin)
        if stdout:
            self.assertEqual(stdout, str(process.stdout))
        else:
            self.assertIsNone(process.stdout)
        if stderr:
            self.assertEqual(stderr, str(process.stderr))
        else:
            self.assertIsNone(process.stderr)
        if expect_exit_status is not None:
            self.assertEqual(0, process.exit_status)

    def test_succeed_with_command_list(self):
        self.test_succeed(['echo', 'something'],
                          stdout='something\n')

    def test_succeed_reading_from_stdout(self):
        self.test_succeed('echo something',
                          stdout='something\n')

    def test_succeed_reading_from_stderr(self):
        self.test_succeed('echo something >&2',
                          stderr='something\n')

    def test_succeed_writing_to_stdin(self):
        self.test_succeed('cat',
                          stdin='some input\n',
                          stdout='some input\n')

    def test_succeed_with_timeout(self):
        self.test_succeed(timeout=30.)

    def test_succeed_with_no_exit_status(self):
        self.test_succeed(command='false', expect_exit_status=None)

    def test_fails(self, command='false', exit_status=None, stdin=None,
                   stdout=None, stderr=None, expect_exit_status=0,
                   **kwargs):
        ex = self.assertRaises(sh.ShellCommandFailed,
                               self.execute,
                               command=command,
                               expect_exit_status=expect_exit_status,
                               stdin=stdin,
                               stdout=bool(stdout),
                               stderr=bool(stderr),
                               **kwargs)
        self.assertEqual(self.expected_command(command), ex.command)
        if stdin:
            self.assertEqual(stdin, ex.stdin)
        else:
            self.assertIsNone(ex.stdin)
        if stdout:
            self.assertEqual(stdout, ex.stdout)
        else:
            self.assertIsNone(ex.stdout)
        if stderr:
            self.assertEqual(stderr, ex.stderr)
        else:
            self.assertIsNone(ex.stderr)
        if exit_status is not None:
            self.assertEqual(exit_status, ex.exit_status)
        else:
            self.assertTrue(ex.exit_status)

    def test_fails_getting_exit_status(self):
        self.test_fails('exit 15', exit_status=15)

    def test_fails_reading_from_stdout(self):
        self.test_fails(command='echo something && false',
                        stdout='something\n')

    def test_fails_reading_from_stderr(self):
        self.test_fails(command='echo something >&2 && false',
                        stderr='something\n')

    def test_fails_writing_to_stdin(self):
        self.test_fails('cat && false',
                        stdin='some input\n',
                        stdout='some input\n')

    def test_fails_with_check_exit_status(self):
        self.test_fails(command='true', expect_exit_status=1, exit_status=0)

    def test_timeout_expires(self, command='sleep 10', timeout=5., stdin=None,
                             stdout=None, stderr=None, **kwargs):
        ex = self.assertRaises(sh.ShellTimeoutExpired,
                               self.execute,
                               command=command,
                               timeout=timeout,
                               stdin=stdin,
                               stdout=bool(stdout),
                               stderr=bool(stderr),
                               **kwargs)
        self.assertEqual(self.expected_command(command), ex.command)
        if stdin:
            self.assertTrue(stdin.startswith(ex.stdin))
        else:
            self.assertIsNone(ex.stdin)
        if stdout:
            self.assertTrue(stdout.startswith(ex.stdout))
        else:
            self.assertIsNone(ex.stdout)
        if stderr:
            self.assertTrue(stderr.startswith(ex.stderr))
        else:
            self.assertIsNone(ex.stderr)
        self.assertEqual(timeout, ex.timeout)

    def execute(self, **kwargs):
        kwargs.setdefault('shell', self.shell)
        kwargs.setdefault('ssh_client', self.ssh_client)
        return sh.execute(**kwargs)

    def expected_command(self, command):
        command = sh.shell_command(command)
        if self.shell:
            command = sh.shell_command(self.shell) + [str(command)]
        return str(command)


class LocalExecuteTest(ExecuteTest):

    def execute(self, **kwargs):
        kwargs.setdefault('shell', self.shell)
        return sh.local_execute(**kwargs)


@keystone.skip_unless_has_keystone_credentials()
class SSHExecuteTest(ExecuteTest):

    server_stack = tobiko.required_setup_fixture(
        stacks.CirrosServerStackFixture)

    @property
    def ssh_client(self):
        return self.server_stack.ssh_client

    def execute(self, **kwargs):
        kwargs.setdefault('shell', self.shell)
        return sh.ssh_execute(ssh_client=self.ssh_client, **kwargs)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr
from oslo_log import log
import testtools

import tobiko
from tobiko.shell import files
from tobiko.shell import ping
from tobiko.shell import ip
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack import topology


LOG = log.getLogger(__name__)


class PortTest(testtools.TestCase):
    """Test Neutron ports"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def test_port_ips(self, ip_version: typing.Optional[int] = None):
        """Checks port IPS has been assigned to server via DHCP protocol"""
        device_ips = set(neutron.list_device_ip_addresses(
            device_id=self.stack.server_id,
            network_id=self.stack.network_stack.network_id,
            enable_dhcp=True,
            ip_version=ip_version))
        if device_ips:
            server_ips = set(ip.list_ip_addresses(
                scope='global', ssh_client=self.stack.ssh_client))
            self.assertEqual(device_ips, device_ips & server_ips)
        elif ip_version:
            self.skipTest(f"No port IPv{ip_version} addresses found")
        else:
            self.skipTest("No port IP addresses found")

    def test_port_network(self):
        self.assertEqual(self.stack.network_stack.network_id,
                         self.stack.port_details['network_id'])

    def test_port_subnets(self):
        """Checks port subnets"""
        port_subnets = [fixed_ip['subnet_id']
                        for fixed_ip in self.stack.port_details['fixed_ips']]
        network_subnets = self.stack.network_stack.network_details['subnets']
        self.assertEqual(set(network_subnets), set(port_subnets))

    def test_ping_subnet_gateways(self):
        """Checks server can ping its gateway IPs"""
        network_id = self.stack.network_stack.network_id
        subnets = neutron.list_subnets(network_id=network_id,
                                       enable_dhcp=True)
        LOG.debug(f"Subnets with DHCP enabled are: {subnets}")
        gateway_ips = [netaddr.IPAddress(subnet['gateway_ip'])
                       for subnet in subnets]
        LOG.debug(f"Gateway IPs are: {gateway_ips}")
        ping.assert_reachable_hosts(gateway_ips,
                                    ssh_client=self.stack.ssh_client)

    def test_ping_port(self, network_id=None, device_id=None, ip_version=None):
        """Checks server can ping its own port"""
        device_ips = neutron.list_device_ip_addresses(
            device_id=device_id or self.stack.server_id,
            network_id=network_id or self.stack.network_stack.network_id,
            enable_dhcp=True, ip_version=ip_version)
        server_ips = ip.list_ip_addresses(scope='global',
                                          ssh_client=self.stack.ssh_client)
        # Remove IPs that hasn't been assigned to server
        port_ips = tobiko.Selection(set(device_ips) & set(server_ips))
        if port_ips:
            ping.assert_reachable_hosts(port_ips,
                                        ssh_client=self.stack.ssh_client)
        elif ip_version:
            self.skipTest(f"No port IPv{ip_version} addresses found")
        else:
            self.skipTest("No port IP addresses found")


# --- Test opening ports on external network ----------------------------------

@stacks.skip_unless_has_external_network
class ExternalPortTest(PortTest):
    """Test Neutron ports"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(
        stacks.CirrosExternalServerStackFixture)


# --- Test la-h3 extension ----------------------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3HAPortTest(PortTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(stacks.L3haServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class CentosServerL3HAPortTest(PortTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(stacks.L3haCentosServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class UbuntuServerL3HAPortTest(PortTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(stacks.L3haUbuntuServerStackFixture)


# --- Port events logging ----------------------------------------------------

class PortLogsStack(stacks.CirrosServerStackFixture):
    pass


@neutron.skip_unless_is_ovs()
class PortLogsTest(testtools.TestCase):

    stack = tobiko.required_setup_fixture(PortLogsStack)

    def setUp(self):
        super(PortLogsTest, self).setUp()
        os_topology = topology.get_openstack_topology()
        self.LOG_FILENAME = os_topology.log_names_mappings[neutron.SERVER]
        self.FILE_DIGGER_CLASS = os_topology.file_digger_class

    def test_nova_port_notification(self):
        pattern = f'Nova.+event.+response.*{self.stack.server_id}'
        log_digger = files.MultihostLogFileDigger(
            filename=self.LOG_FILENAME,
            file_digger_class=self.FILE_DIGGER_CLASS,
            sudo=True)
        for node in topology.list_openstack_nodes(group='controller'):
            log_digger.add_host(hostname=node.hostname,
                                ssh_client=node.ssh_client)
        log_digger.find_lines(pattern=pattern)

        nova.shutoff_server(self.stack.server_id)
        nova.activate_server(self.stack.server_id)

        new_lines = log_digger.find_new_lines(pattern=pattern)

        plugged_events = [
            (hostname, line)
            for hostname, line in new_lines
            if 'network-vif-plugged' in line and self.stack.port_id in line]
        self.assertEqual(1, len(plugged_events), new_lines)

        unplugged_events = [
            (hostname, line)
            for hostname, line in new_lines
            if 'network-vif-unplugged' in line and self.stack.port_id in line]
        self.assertEqual(1, len(unplugged_events), new_lines)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import inspect

from keystoneauth1 import session as _session
import mock

import tobiko
from tobiko.openstack import _client
from tobiko.openstack import keystone
from tobiko.tests.unit import openstack


CLIENT = object()


class ClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session=None):
        return CLIENT


MockSession = mock.create_autospec(_session.Session)


class OpenstackClientFixtureTest(openstack.OpenstackTest):

    def create_client(self, session=None):
        return ClientFixture(session=session)

    def test_init(self, session=None):
        client = self.create_client(session=session)
        self.assertIs(session or None, client.session)

    def test_init_with_credentials(self):
        self.test_init(session=MockSession)

    def test_init_with_credentials_fixture(self):
        self.test_init(session=keystone.KeystoneSessionFixture())

    def test_init_with_credentials_fixture_type(self):
        self.test_init(session=keystone.KeystoneSessionFixture)

    def test_setup(self, session=None):
        client = self.create_client(session=session)
        client.setUp()
        if session:
            if tobiko.is_fixture(session):
                if inspect.isclass(session):
                    session = tobiko.get_fixture(session)
                self.assertIs(session.session, client.session)
            else:
                self.assertIs(session, client.session)
        else:
            self.assertIs(keystone.get_keystone_session(), client.session)

    def test_setup_with_session(self):
        self.test_setup(session=MockSession())

    def test_setup_with_session_fixture(self):
        self.test_setup(session=keystone.KeystoneSessionFixture())

    def test_setup_with_session_fixture_type(self):
        self.test_setup(session=keystone.KeystoneSessionFixture)


class ClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return ClientFixture(session=session)


class OpenstackClientManagerTest(openstack.OpenstackTest):

    def setUp(self):
        super(OpenstackClientManagerTest, self).setUp()
        self.patch(keystone, 'get_keystone_session',
                   return_value=MockSession())

    def test_init(self):
        manager = ClientManager()
        self.assertEqual({}, manager.clients)

    def test_get_client(self, session=None, shared=True):
        manager = ClientManager()
        client1 = manager.get_client(session=session, shared=shared)
        client2 = manager.get_client(session=session, shared=shared)
        if shared:
            self.assertIs(client1, client2)
        else:
            self.assertIsNot(client1, client2)

    def test_get_client_with_not_shared(self):
        self.test_get_client(shared=False)

    def test_get_client_with_session(self):
        self.test_get_client(session=MockSession())

    def test_get_client_with_session_fixture(self):
        self.test_get_client(session=keystone.KeystoneSessionFixture())

    def test_get_client_with_session_fixture_type(self):
        self.test_get_client(session=keystone.KeystoneSessionFixture)

    def test_get_client_with_init_client(self):
        init_client = mock.MagicMock(return_value=CLIENT)
        manager = _client.OpenstackClientManager()
        session = MockSession()
        client = manager.get_client(session=session,
                                    init_client=init_client)
        self.assertIs(CLIENT, client)
        init_client.assert_called_once_with(session=session)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import contextlib
import getpass
import os
import socket
import subprocess
import typing  # noqa

import netaddr
from oslo_log import log
import paramiko
import six

import tobiko
from tobiko.shell.ssh import _config
from tobiko.shell.ssh import _command


LOG = log.getLogger(__name__)


def valid_hostname(value):
    hostname = str(value)
    if not hostname:
        message = "Invalid hostname: {!r}".format(hostname)
        raise ValueError(message)
    return hostname


def valid_port(value):
    port = int(value)
    if port <= 0 or port > 65535:
        message = "Invalid port number: {!r}".format(port)
        raise ValueError(message)
    return port


def valid_path(value):
    return os.path.abspath(os.path.expanduser(value))


def positive_float(value):
    value = float(value)
    if value <= 0.:
        message = "{!r} is not positive".format(value)
        raise ValueError(message)
    return value


def positive_int(value):
    value = int(value)
    if value <= 0:
        message = "{!r} is not positive".format(value)
        raise ValueError(message)
    return value


def get_key_filename(value):
    if isinstance(value, six.string_types):
        value = [value]
    key_filename = [tobiko.tobiko_config_path(v) for v in value]
    return [f
            for f in key_filename
            if os.path.isfile(f)]


SSH_CONNECT_PARAMETERS = {
    #: The server to connect to
    'hostname': valid_hostname,

    #: The server port to connect to
    'port': valid_port,

    #: The username to authenticate as (defaults to the current local username)
    'username': str,

    #: Used for password authentication; is also used for private key
    #: decryption if passphrase is not given
    'password': str,

    #: Used for decrypting private keys
    'passphrase': str,

    #: The filename, or list of filenames, of optional private key(s) and/or
    #: certs to try for authentication
    'key_filename': get_key_filename,

    #: An optional timeout (in seconds) for the TCP connect
    'timeout': positive_float,

    #: Set to False to disable connecting to the SSH agent
    'allow_agent': bool,

    #: Set to False to disable searching for discoverable private key files in
    #: ~/.ssh/
    'look_for_keys': bool,

    #: Set to True to turn on compression
    'compress': bool,

    #: True if you want to use GSS-API authentication
    'gss_auth': bool,

    #: Perform GSS-API Key Exchange and user authentication
    'gss_kex': bool,

    #: Delegate GSS-API client credentials or not
    'gss_deleg_creds': bool,

    #: The targets name in the kerberos database. default: hostname
    'gss_host': str,

    #: Indicates whether or not the DNS is trusted to securely canonicalize the
    #: name of the host being connected to (default True).
    'gss_trust_dns': bool,

    #: An optional timeout (in seconds) to wait for the SSH banner to be
    #: presented.
    'banner_timeout': positive_float,

    #: An optional timeout (in seconds) to wait for an authentication response
    'auth_timeout': positive_float,

    #: Number of connection attempts to be tried before timeout
    'connection_attempts': positive_int,

    #: Minimum amount of time to wait between two connection attempts
    'connection_interval': positive_float,

    #: Amount of time before timing our connection attempt
    'connection_timeout': positive_float,

    #: Command to be executed to open proxy sock
    'proxy_command': str,
}


def gather_ssh_connect_parameters(source=None, destination=None, schema=None,
                                  remove_from_schema=False, **kwargs):
    if schema is None:
        assert not remove_from_schema
        schema = SSH_CONNECT_PARAMETERS
    parameters = {}

    if source:
        # gather from object
        if isinstance(source, collections.Mapping):
            parameters.update(_items_from_mapping(mapping=source,
                                                  schema=schema))
        else:
            parameters.update(_items_from_object(obj=source,
                                                 schema=schema))

    if kwargs:
        # gather from kwargs
        parameters.update(_items_from_mapping(mapping=kwargs, schema=schema))
        kwargs = exclude_mapping_items(kwargs, schema)
        if kwargs:
            message = "Invalid SSH connection parameters: {!r}".format(kwargs)
            raise ValueError(message)

    if remove_from_schema and parameters:
        # update schema
        for name in parameters:
            del schema[name]

    if destination is not None:
        destination.update(parameters)
    return parameters


def _items_from_mapping(mapping, schema):
    for name, init in schema.items():
        value = mapping.get(name)
        if value is not None:
            yield name, init(value)


def _items_from_object(obj, schema):
    for name, init in schema.items():
        value = getattr(obj, name, None)
        if value is not None:
            yield name, init(value)


def exclude_mapping_items(mapping, exclude):
    """Exclude parameters that are already in target dictionary"""
    return {key: value
            for key, value in mapping.items()
            if key not in exclude}


class SSHConnectFailure(tobiko.TobikoException):
    message = "Failed to login to {login}\n{cause}"


class SSHClientFixture(tobiko.SharedFixture):

    host = None
    port = 22
    username = getpass.getuser()

    client = None

    default = tobiko.required_setup_fixture(_config.SSHDefaultConfigFixture)
    config_files = None
    host_config = None
    global_host_config = None

    proxy_client = None
    proxy_sock = None
    connect_parameters = None
    schema = SSH_CONNECT_PARAMETERS

    def __init__(self, host=None, proxy_client=None, host_config=None,
                 config_files=None, schema=None, **kwargs):
        super(SSHClientFixture, self).__init__()
        if host:
            self.host = host
        if proxy_client:
            self.proxy_client = proxy_client
        if host_config:
            self.host_config = host_config
        if config_files:
            self.config_files = config_files

        self.schema = schema = dict(schema or self.schema)
        self._connect_parameters = gather_ssh_connect_parameters(
            schema=schema, **kwargs)
        self._forwarders = []

    def setup_fixture(self):
        self.setup_connect_parameters()
        self.setup_ssh_client()

    def setup_global_host_config(self):
        self.global_host_config = config = _config.ssh_host_config(
            host=self.host, config_files=self.config_files)
        return config

    def setup_connect_parameters(self):
        """Fill connect parameters dict

        Get parameters values from below sources:
        - parameters passed to class constructor
        - parameters got from ~/.ssh/config and tobiko.conf
        - parameters got from fixture object attributes
        """
        self.setup_global_host_config()
        self.connect_parameters = parameters = self.get_connect_parameters()
        return parameters

    def get_connect_parameters(self, schema=None):
        schema = dict(schema or self.schema)
        parameters = {}
        self.gather_initial_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        self.gather_global_host_config_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        self.gather_host_config_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        self.gather_default_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        return parameters

    def gather_initial_connect_parameters(self, **kwargs):
        if self._connect_parameters:
            gather_ssh_connect_parameters(
                source=self._connect_parameters, **kwargs)

    def gather_host_config_connect_parameters(self, **kwargs):
        if self.host_config:
            gather_ssh_connect_parameters(
                source=self.host_config.connect_parameters, **kwargs)

    def gather_global_host_config_connect_parameters(self, **kwargs):
        if self.global_host_config:
            gather_ssh_connect_parameters(
                source=self.global_host_config.connect_parameters, **kwargs)

    def gather_default_connect_parameters(self, **kwargs):
        return gather_ssh_connect_parameters(source=self, **kwargs)

    def setup_ssh_client(self):
        self.client, self.proxy_sock = ssh_connect(
            proxy_client=self.proxy_client,
            **self.connect_parameters)
        self.addCleanup(self.cleanup_ssh_client)
        if self.proxy_sock:
            self.addCleanup(self.cleanup_proxy_sock)
        for forwarder in self._forwarders:
            self.useFixture(forwarder)

    def cleanup_ssh_client(self):
        client = self.client
        self.client = None
        if client:
            LOG.debug(f"Closing SSH connection... ({self.details})")
            try:
                client.close()
            except Exception:
                LOG.exception("Error closing SSH connection. "
                              f"({self.details})")
            else:
                LOG.debug(f"SSH connection closed. ({self.details})")

    def cleanup_proxy_sock(self):
        proxy_sock = self.proxy_sock
        self.proxy_sock = None
        if proxy_sock:
            LOG.debug(f"Closing SSH proxy sock... ({self.details})")
            try:
                proxy_sock.close()
            except Exception:
                LOG.exception("Error closing proxy socket. "
                              f"({self.details})")
            else:
                LOG.debug(f"SSH proxy sock closed. ({self.details})")

    @contextlib.contextmanager
    def use_connect_parameters(self, **kwargs):
        if kwargs:
            restore_parameters = dict(self._connect_parameters)
            gather_ssh_connect_parameters(schema=self.schema,
                                          destination=self._connect_parameters,
                                          **kwargs)
            self.connect_parameters = self.get_connect_parameters()
        else:
            restore_parameters = None
        try:
            yield
        finally:
            if restore_parameters is not None:
                self._connect_parameters = restore_parameters
                self.connect_parameters = self.get_connect_parameters()

    def connect(self, retry_count: typing.Optional[int] = 2,
                retry_timeout: tobiko.Seconds = None,
                retry_interval: tobiko.Seconds = None, **ssh_parameters):
        """Ensures it is connected to remote SSH server
        """
        with self.use_connect_parameters(**ssh_parameters):
            # This retry is mostly intended to ensure connection is
            # reestablished in case it is lost
            for attempt in tobiko.retry(count=retry_count,
                                        timeout=retry_timeout,
                                        interval=retry_interval):
                LOG.debug(f"Ensuring SSH connection (attempt={attempt})")
                connected = False
                try:
                    client = tobiko.setup_fixture(self).client
                    # For any reason at this point client could
                    # be None: force fixture cleanup
                    if check_ssh_connection(client):
                        LOG.debug("SSH connection is safe to use "
                                  f"(attempt={attempt})")
                        connected = True
                        return client
                    else:
                        LOG.warning("SSH connection is not safe to use "
                                    f"(attempt={attempt})")
                except Exception:
                    attempt.check_limits()
                    LOG.exception(f"Failed connecting to '{self.login}' "
                                  f"(attempt={attempt})")
                finally:
                    if not connected:
                        self.close()

    def close(self):
        """Ensures it is disconnected from remote SSH server
        """
        try:
            tobiko.cleanup_fixture(self)
        except Exception:
            LOG.exception(f"Failed closing SSH connection to '{self.login}'")

    def get_ssh_command(self, host=None, username=None, port=None,
                        command=None, config_files=None, host_config=None,
                        proxy_command=None, key_filename=None, **options):
        connect_parameters = self.setup_connect_parameters()
        host = host or connect_parameters.get('hostname')
        username = username or connect_parameters.get('username')
        port = port or connect_parameters.get('port')
        config_files = config_files or connect_parameters.get('config_files')
        if not host_config:
            _host_config = self.setup_global_host_config()
            if hasattr(_host_config, 'host_config'):
                _host_config = host_config
        key_filename = key_filename or connect_parameters.get('key_filename')
        proxy_command = (proxy_command or
                         connect_parameters.get('proxy_command'))
        if not proxy_command and self.proxy_client:
            proxy_command = self.proxy_client.get_ssh_command()

        return _command.ssh_command(host=host,
                                    username=username,
                                    port=port,
                                    command=command,
                                    config_files=config_files,
                                    host_config=host_config,
                                    proxy_command=proxy_command,
                                    key_filename=key_filename,
                                    **options)

    @property
    def login(self):
        parameters = self.setup_connect_parameters()
        return _command.ssh_login(hostname=parameters['hostname'],
                                  username=parameters['username'],
                                  port=parameters['port'])

    @property
    def hostname(self):
        parameters = self.setup_connect_parameters()
        return parameters['hostname']

    @property
    def details(self):
        return f"login='{self.login}'"

    def __repr__(self):
        return f"SSHClientFixture <{self.details}>"


UNDEFINED_CLIENT = 'UNDEFINED_CLIENT'


class SSHClientManager(object):

    default = tobiko.required_setup_fixture(_config.SSHDefaultConfigFixture)

    def __init__(self):
        self.clients = {}

    def get_client(self, host, hostname=None, username=None, port=None,
                   proxy_jump=None, host_config=None, config_files=None,
                   proxy_client=None, **connect_parameters) -> \
            SSHClientFixture:
        if isinstance(host, netaddr.IPAddress):
            host = str(host)

        if host_config:
            hostname = hostname or host_config.hostname
            port = port or host_config.port
            username = username or host_config.username

        global_host_config = _config.ssh_host_config(host=host,
                                                     config_files=config_files)
        hostname = hostname or global_host_config.hostname
        port = port or global_host_config.port
        username = username or global_host_config.username

        host_key = hostname, port, username, proxy_jump
        existing_client = self.clients.get(host_key)
        if isinstance(existing_client, SSHClientFixture):
            return existing_client

        # Put a placeholder to avoid infinite recursive lookup
        if existing_client is UNDEFINED_CLIENT:
            raise RuntimeError('Recursive SSH proxy client definition')
        self.clients[host_key] = UNDEFINED_CLIENT

        proxy_client = proxy_client or self.get_proxy_client(
            host=host, proxy_jump=proxy_jump, config_files=config_files)
        self.clients[host_key] = new_client = SSHClientFixture(
            host=host, hostname=hostname, port=port, username=username,
            proxy_client=proxy_client, host_config=host_config,
            **connect_parameters)
        return new_client

    def get_proxy_client(self, host=None, proxy_jump=None, host_config=None,
                         config_files=None):
        if isinstance(proxy_jump, SSHClientFixture):
            return proxy_jump
        host_config = host_config or _config.ssh_host_config(
            host=host, config_files=config_files)
        proxy_jump = host_config.proxy_jump
        return proxy_jump and self.get_client(proxy_jump) or None


CLIENTS = SSHClientManager()


def ssh_client(host, port=None, username=None, proxy_jump=None,
               host_config=None, config_files=None, manager=None,
               **connect_parameters) -> SSHClientFixture:
    manager = manager or CLIENTS
    return manager.get_client(host=host, port=port, username=username,
                              proxy_jump=proxy_jump, host_config=host_config,
                              config_files=config_files,
                              **connect_parameters)


def ssh_connect(hostname, username=None, port=None, connection_interval=None,
                connection_attempts=None, connection_timeout=None,
                proxy_command=None, proxy_client=None, key_filename=None,
                **parameters):
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.WarningPolicy())
    login = _command.ssh_login(hostname=hostname, username=username, port=port)

    if key_filename:
        # Ensures we try enough times to try all keys
        tobiko.check_valid_type(key_filename, list)
        connection_attempts = max(connection_attempts or 1,
                                  len(key_filename),
                                  1)

    for attempt in tobiko.retry(count=connection_attempts,
                                timeout=connection_timeout,
                                interval=connection_interval,
                                default_count=60,
                                default_timeout=300.,
                                default_interval=5.):
        LOG.debug(f"Logging in to '{login}'...\n"
                  f"  - parameters: {parameters}\n"
                  f"  - attempt: {attempt.details}\n")

        try:
            proxy_sock = ssh_proxy_sock(
                hostname=hostname,
                port=port,
                command=proxy_command,
                client=proxy_client,
                timeout=connection_timeout,
                connection_attempts=1,
                connection_interval=connection_interval)
            client.connect(hostname=hostname,
                           username=username,
                           port=port,
                           sock=proxy_sock,
                           key_filename=key_filename,
                           **parameters)
        except ValueError as ex:
            attempt.check_limits()
            if (str(ex) == 'q must be exactly 160, 224, or 256 bits long' and
                    len(key_filename) > 1):
                # Must try without the first key
                LOG.debug(f"Retry connecting with the next key: {ex}")
                key_filename = key_filename[1:]
                continue
            else:
                raise
        except (EOFError, socket.error, socket.timeout,
                paramiko.SSHException) as ex:
            attempt.check_limits()
            LOG.debug(f"Error logging in to '{login}': {ex}")
        else:
            LOG.debug(f"Successfully logged in to '{login}'")
            return client, proxy_sock


def ssh_proxy_sock(hostname=None, port=None, command=None, client=None,
                   source_address=None, timeout=None,
                   connection_attempts=None, connection_interval=None):
    if not command:
        if client:
            # I need a command to execute with proxy client
            options = []
            if source_address:
                options += ['-s', str(source_address)]
            command = ['nc'] + options + ['{hostname!s}', '{port!s}']
        elif not command:
            # Proxy sock is not required
            return None

    # Apply connect parameters to proxy command
    if not isinstance(command, six.string_types):
        command = subprocess.list2cmdline(command)
    if hostname:
        command = command.format(hostname=hostname, port=(port or 22))
    if client:
        if isinstance(client, SSHClientFixture):
            # Connect to proxy server
            client = client.connect(connection_timeout=timeout,
                                    connection_attempts=connection_attempts,
                                    connection_interval=connection_interval)
        elif not isinstance(client, paramiko.SSHClient):
            message = "Object {!r} is not an SSHClient".format(client)
            raise TypeError(message)

        # Open proxy channel
        LOG.debug("Execute proxy command with proxy client %r: %r",
                  client, command)
        sock = client.get_transport().open_session(timeout=timeout)
        sock.exec_command(command)
    else:
        LOG.debug("Execute proxy command on local host: %r", command)
        sock = paramiko.ProxyCommand(command)

    return sock


def ssh_proxy_client(manager=None, host=None, host_config=None,
                     config_files=None):
    manager = manager or CLIENTS
    return manager.get_proxy_client(host=host,
                                    host_config=host_config,
                                    config_files=config_files)


def check_ssh_connection(client):
    if client:
        transport = client.get_transport()
        if transport.is_authenticated():
            # Send a keep-alive message
            transport.send_ignore()
            return True
    return False



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import os
import tempfile
import typing  # noqa

import yaml

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack.keystone import _clouds_file
from tobiko.tests.unit import openstack
from tobiko.tests.unit.openstack.keystone import test_credentials


def make_clouds_content(cloud_name, api_version=None, auth=None):
    content = {}
    if api_version is not None:
        content['identity_api_version'] = api_version
    if auth is not None:
        content['auth'] = auth
    return {'clouds': {cloud_name: content}}


class CloudsFileFixture(tobiko.SharedFixture):

    cloud_name = None  # type: str
    api_version = None  # type: str
    auth = None  # type: typing.Dict[str, typing.Any]
    clouds_content = None
    clouds_file = None
    suffix = '.yaml'
    create_file = True

    def __init__(self, cloud_name=None, api_version=None, auth=None,
                 clouds_file=None, suffix=None, create_file=None,
                 clouds_content=None):
        super(CloudsFileFixture, self).__init__()
        if cloud_name is not None:
            self.cloud_name = cloud_name
        if api_version is not None:
            self.api_version = api_version
        if auth is not None:
            self.auth = auth
        if clouds_file is not None:
            self.clouds_file = clouds_file
        if suffix is not None:
            self.suffix = suffix
        if create_file is not None:
            self.create_file = create_file
        if clouds_content is not None:
            self.clouds_content = clouds_content

    def setup_fixture(self):
        clouds_content = self.clouds_content
        if clouds_content is None:
            self.clouds_content = clouds_content = make_clouds_content(
                cloud_name=self.cloud_name, api_version=self.api_version,
                auth=self.auth)

        if self.create_file:
            clouds_file = self.clouds_file
            if clouds_file is None:
                fd, clouds_file = tempfile.mkstemp(suffix=self.suffix)
                self.addCleanup(os.remove, clouds_file)
                self.clouds_file = clouds_file
                clouds_stream = os.fdopen(fd, 'wt')
            else:
                clouds_stream = os.open(clouds_file, 'wt')

            try:
                if self.suffix in _clouds_file.JSON_SUFFIXES:
                    json.dump(clouds_content, clouds_stream)
                elif self.suffix in _clouds_file.YAML_SUFFIXES:
                    yaml.safe_dump(clouds_content, clouds_stream)
            finally:
                clouds_stream.close()


class V2CloudsFileFixture(CloudsFileFixture):
    cloud_name = 'V2-TEST_CLOUD'
    auth = test_credentials.V2_PARAMS


class V3CloudsFileFixture(CloudsFileFixture):
    cloud_name = 'V3-TEST_CLOUD'
    auth = test_credentials.V3_PARAMS


class CloudsFileKeystoneCredentialsFixtureTest(openstack.OpenstackTest):

    config = tobiko.required_setup_fixture(
        _clouds_file.DefaultCloudsFileConfig)

    def test_init(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture()
        self.assertEqual(self.config.cloud_name, fixture.cloud_name)
        self.assertIsNone(fixture.clouds_content)
        self.assertIsNone(fixture.clouds_file)
        self.assertEqual(self.config.clouds_files, fixture.clouds_files)

    def test_init_with_cloud_name(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name')
        self.assertEqual('cloud-name', fixture.cloud_name)

    def test_init_with_clouds_content(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_content={})
        self.assertEqual({}, fixture.clouds_content)

    def test_init_with_clouds_file(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_file='cloud-file')
        self.assertEqual('cloud-file', fixture.clouds_file)

    def test_init_with_clouds_files(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_files=['a', 'b', 'd'])
        self.assertEqual(['a', 'b', 'd'], fixture.clouds_files)

    def test_setup_from_default_clouds_files(self):
        file_fixture = self.useFixture(V3CloudsFileFixture())
        self.patch(self.config, 'clouds_files',
                   ['/a', file_fixture.clouds_file, '/c'])
        credentials_fixture = self.useFixture(
            keystone.CloudsFileKeystoneCredentialsFixture(
                cloud_name=file_fixture.cloud_name))
        self.assertEqual(file_fixture.clouds_content,
                         credentials_fixture.clouds_content)
        self.assertEqual(test_credentials.V3_PARAMS,
                         credentials_fixture.credentials.to_dict())

    def test_setup_from_json(self):
        file_fixture = self.useFixture(V3CloudsFileFixture(suffix='.json'))
        credentials_fixture = self.useFixture(
            keystone.CloudsFileKeystoneCredentialsFixture(
                cloud_name=file_fixture.cloud_name,
                clouds_file=file_fixture.clouds_file))
        self.assertEqual(file_fixture.clouds_content,
                         credentials_fixture.clouds_content)
        self.assertEqual(test_credentials.V3_PARAMS,
                         credentials_fixture.credentials.to_dict())

    def test_setup_from_yaml(self):
        file_fixture = self.useFixture(V3CloudsFileFixture(suffix='.yaml'))
        credentials_fixture = self.useFixture(
            keystone.CloudsFileKeystoneCredentialsFixture(
                cloud_name=file_fixture.cloud_name,
                clouds_file=file_fixture.clouds_file))
        self.assertEqual(file_fixture.clouds_content,
                         credentials_fixture.clouds_content)
        self.assertEqual(test_credentials.V3_PARAMS,
                         credentials_fixture.credentials.to_dict())

    def test_setup_from_yml(self):
        file_fixture = self.useFixture(V3CloudsFileFixture(suffix='.yml'))
        credentials_fixture = self.useFixture(
            keystone.CloudsFileKeystoneCredentialsFixture(
                cloud_name=file_fixture.cloud_name,
                clouds_file=file_fixture.clouds_file))
        self.assertEqual(file_fixture.clouds_content,
                         credentials_fixture.clouds_content)
        self.assertEqual(test_credentials.V3_PARAMS,
                         credentials_fixture.credentials.to_dict())

    def test_setup_v2_credentials(self):
        file_fixture = self.useFixture(V2CloudsFileFixture())
        credentials_fixture = self.useFixture(
            keystone.CloudsFileKeystoneCredentialsFixture(
                cloud_name=file_fixture.cloud_name,
                clouds_file=file_fixture.clouds_file))
        self.assertEqual(file_fixture.clouds_content,
                         credentials_fixture.clouds_content)
        self.assertEqual(test_credentials.V2_PARAMS,
                         credentials_fixture.credentials.to_dict())

    def test_setup_with_cloud_name(self):
        file_fixture = self.useFixture(V3CloudsFileFixture())
        credentials_fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name',
            clouds_file=file_fixture.clouds_file)
        ex = self.assertRaises(ValueError, tobiko.setup_fixture,
                               credentials_fixture)
        self.assertEqual("No such cloud with name 'cloud-name' in file " +
                         repr(file_fixture.clouds_file), str(ex))

    def test_setup_with_cloud_name_from_env(self):
        self.patch(self.config, 'cloud_name', None)

        file_fixture = self.useFixture(V2CloudsFileFixture())

        self.patch(os, 'environ', {'OS_CLOUD': file_fixture.cloud_name})
        credentials_fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_file=file_fixture.clouds_file)

        self.assertIsNone(credentials_fixture.cloud_name)
        tobiko.setup_fixture(credentials_fixture)
        self.assertEqual(file_fixture.cloud_name,
                         credentials_fixture.cloud_name)

    def test_setup_with_empty_cloud_name(self):
        file_fixture = self.useFixture(V2CloudsFileFixture())
        credentials_fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_file=file_fixture.clouds_file,
            cloud_name='')

        self.assertIsNone(credentials_fixture.credentials)
        self.assertEqual('', credentials_fixture.cloud_name)
        tobiko.setup_fixture(credentials_fixture)
        self.assertIsNone(credentials_fixture.credentials)
        self.assertEqual('', credentials_fixture.cloud_name)

    def test_setup_with_empty_cloud_name_from_env(self):
        self.patch(self.config, 'cloud_name', None)

        file_fixture = self.useFixture(V2CloudsFileFixture())
        self.patch(os, 'environ', {'OS_CLOUD': ''})
        credentials_fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_file=file_fixture.clouds_file)

        self.assertIsNone(credentials_fixture.credentials)
        self.assertIsNone(credentials_fixture.cloud_name)
        tobiko.setup_fixture(credentials_fixture)
        self.assertIsNone(credentials_fixture.credentials)
        self.assertIsNone(credentials_fixture.cloud_name)

    def test_setup_with_no_cloud_name(self):
        self.patch(self.config, 'cloud_name', None)

        file_fixture = self.useFixture(V2CloudsFileFixture())
        credentials_fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_file=file_fixture.clouds_file)

        self.assertIsNone(credentials_fixture.credentials)
        self.assertIsNone(credentials_fixture.cloud_name)
        tobiko.setup_fixture(credentials_fixture)
        self.assertIsNone(credentials_fixture.credentials)
        self.assertIsNone(credentials_fixture.cloud_name)

    def test_setup_with_no_clouds_section(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name', clouds_content={'other_data': None},
            clouds_file='clouds-file')
        ex = self.assertRaises(ValueError, tobiko.setup_fixture, fixture)
        self.assertEqual('cloud-name', fixture.cloud_name)
        self.assertEqual({'other_data': None}, fixture.clouds_content)
        self.assertEqual("'clouds' section not found in clouds file "
                         "'clouds-file'", str(ex))

    def test_setup_with_empty_clouds_content(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name', clouds_content={})
        ex = self.assertRaises(ValueError, tobiko.setup_fixture, fixture)
        self.assertEqual('cloud-name', fixture.cloud_name)
        self.assertEqual({}, fixture.clouds_content)
        self.assertEqual('Invalid clouds file content: {}', str(ex))

    def test_setup_with_no_auth(self):
        clouds_content = make_clouds_content('cloud-name')
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name',
            clouds_content=clouds_content,
            clouds_file='cloud-file')
        ex = self.assertRaises(ValueError, tobiko.setup_fixture, fixture)
        self.assertEqual('cloud-name', fixture.cloud_name)
        self.assertEqual(
            "No such 'auth' section in cloud file 'cloud-file' for cloud "
            "name 'cloud-name'", str(ex))

    def test_setup_with_no_auth_url(self):
        clouds_content = make_clouds_content('cloud-name', auth={})
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name',
            clouds_content=clouds_content,
            clouds_file='cloud-file')
        ex = self.assertRaises(ValueError, tobiko.setup_fixture, fixture)
        self.assertEqual('cloud-name', fixture.cloud_name)
        self.assertEqual(
            "No such 'auth_url' in file 'cloud-file' for cloud name "
            "'cloud-name'", str(ex))

    def test_setup_without_clouds_file(self):
        self.patch(self.config, 'clouds_files', ['/a', '/b', '/c'])
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            cloud_name='cloud-name')
        ex = self.assertRaises(_clouds_file.CloudsFileNotFoundError,
                               tobiko.setup_fixture, fixture)
        self.assertEqual('cloud-name', fixture.cloud_name)
        self.assertEqual("No such clouds file(s): /a, /b, /c", str(ex))

    def test_setup_with_non_existing_clouds_file(self):
        fixture = keystone.CloudsFileKeystoneCredentialsFixture(
            clouds_file='/a.yaml',
            cloud_name='cloud-name')
        ex = self.assertRaises(_clouds_file.CloudsFileNotFoundError,
                               tobiko.setup_fixture, fixture)
        self.assertEqual("No such clouds file(s): /a.yaml", str(ex))



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import mock

from tobiko import docker
from tobiko.tests import unit


# This class help us to emulate tobiko.shell.sh.execute
# to simulate errors or not and test the tobiko.docker._shell module
class FakeProcess:
    exit_status = 0
    stdout = ''
    stderr = ''


class TestShell(unit.TobikoUnitTest):

    def setUp(self):
        super(TestShell, self).setUp()
        self.fproc = FakeProcess()

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_docker_urls_single(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = 'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd -H fd:// ' \
                            '--containerd=/run/containerd/containerd.sock'
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.discover_docker_urls(), ['fd://'])

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_docker_urls_multiple(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = 'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd -H fd:// ' \
                            '-H fd:// ' \
                            '--containerd=/run/containerd/containerd.sock'
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.discover_docker_urls(), ['fd://', 'fd://'])

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_docker_urls_separated(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = 'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd -H fd:// ' \
                            '--containerd=/run/containerd/containerd.sock ' \
                            '-H boom'
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.discover_docker_urls(), ['fd://', 'boom'])

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_docker_urls_many_daemons(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = 'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd ' \
                            '-H fd://  -H boom ' \
                            '--containerd=/run/containerd/containerd.sock\n' \
                            'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd -H foo ' \
                            '--containerd=/run/containerd/containerd.sock ' \
                            '-H bar'
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.discover_docker_urls(),
                         ['fd://', 'boom', 'foo', 'bar'])

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_docker_urls_no_daemons(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = ''
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertRaises(docker.DockerUrlNotFoundError,
                          docker.discover_docker_urls)

    @mock.patch('tobiko.shell.sh.execute')
    def test_is_docker_running(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = 'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd ' \
                            '-H fd://  -H boom ' \
                            '--containerd=/run/containerd/containerd.sock\n' \
                            'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd -H foo ' \
                            '--containerd=/run/containerd/containerd.sock ' \
                            '-H bar'
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.is_docker_running(), True)

    @mock.patch('tobiko.shell.sh.execute')
    def test_is_docker_running_without_results(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = ''
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.is_docker_running(), False)

    @mock.patch('tobiko.shell.sh.execute')
    def test_is_docker_running_exit_1(self, mock_execute):
        self.fproc.exit_status = 1
        self.fproc.stdout = 'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd ' \
                            '-H fd://  -H boom ' \
                            '--containerd=/run/containerd/containerd.sock\n' \
                            'root      1999  0.0  0.2 1456736 85272 ?' \
                            'Ssl  15:16   0:01 /usr/bin/dockerd -H foo ' \
                            '--containerd=/run/containerd/containerd.sock ' \
                            '-H bar'
        self.fproc.stderr = ''
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.is_docker_running(), False)

    @mock.patch('tobiko.shell.sh.execute')
    def test_is_docker_running_error_exist_but_exit_0(self, mock_execute):
        self.fproc.exit_status = 0
        self.fproc.stdout = ''
        self.fproc.stderr = 'boom'
        mock_execute.return_value = self.fproc
        self.assertEqual(docker.is_docker_running(), False)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa

from tobiko import _exception


T = typing.TypeVar('T')


class Selection(list, typing.Generic[T]):

    def with_attributes(self, **attributes):
        return self.create(
            filter_by_attributes(self, exclude=False, **attributes))

    def without_attributes(self, **attributes):
        return self.create(
            filter_by_attributes(self, exclude=True, **attributes))

    def with_items(self, **items):
        return self.create(filter_by_items(self, exclude=False, **items))

    def without_items(self, **items):
        return self.create(filter_by_items(self, exclude=True, **items))

    @classmethod
    def create(cls, objects: typing.Iterable[T]):
        return cls(objects)

    @property
    def first(self) -> T:
        if self:
            return self[0]
        else:
            raise ObjectNotFound()

    @property
    def unique(self) -> T:
        if len(self) > 1:
            raise MultipleObjectsFound(list(self))
        else:
            return self.first

    def __repr__(self):
        return '{!s}({!r})'.format(type(self).__name__, list(self))


def select(objects: typing.Iterable[T]) -> Selection[T]:
    return Selection.create(objects)


def filter_by_attributes(objects, exclude=False, **attributes):
    exclude = bool(exclude)
    for obj in objects:
        for key, value in attributes.items():
            matching = value == getattr(obj, key)
            if matching is exclude:
                break
        else:
            yield obj


def filter_by_items(dictionaries, exclude=False, **items):
    exclude = bool(exclude)
    for dictionary in dictionaries:
        for key, value in items.items():
            matching = value == dictionary[key]
            if matching is exclude:
                break
        else:
            yield dictionary


class ObjectNotFound(_exception.TobikoException):
    "Object not found"


class MultipleObjectsFound(_exception.TobikoException):
    "Multiple objects found: {objects!r}"



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.nova import _client
from tobiko.openstack.nova import _cloud_init
from tobiko.openstack.nova import _hypervisor
from tobiko.openstack.nova import _server
from tobiko.openstack.nova import _service


CLIENT_CLASSES = _client.CLIENT_CLASSES
get_console_output = _client.get_console_output
get_nova_client = _client.get_nova_client
get_server = _client.get_server
find_hypervisor = _client.find_hypervisor
find_server = _client.find_server
find_service = _client.find_service
HasNovaClientMixin = _client.HasNovaClientMixin
list_hypervisors = _client.list_hypervisors
list_servers = _client.list_servers
list_services = _client.list_services
nova_client = _client.nova_client
NovaClientFixture = _client.NovaClientFixture
wait_for_server_status = _client.wait_for_server_status
WaitForServerStatusError = _client.WaitForServerStatusError
WaitForServerStatusTimeout = _client.WaitForServerStatusTimeout
shutoff_server = _client.shutoff_server
activate_server = _client.activate_server
migrate_server = _client.migrate_server
confirm_resize = _client.confirm_resize

WaitForCloudInitTimeoutError = _cloud_init.WaitForCloudInitTimeoutError
cloud_config = _cloud_init.cloud_config
get_cloud_init_status = _cloud_init.get_cloud_init_status
user_data = _cloud_init.user_data
wait_for_cloud_init_done = _cloud_init.wait_for_cloud_init_done
wait_for_cloud_init_status = _cloud_init.wait_for_cloud_init_status

skip_if_missing_hypervisors = _hypervisor.skip_if_missing_hypervisors
get_same_host_hypervisors = _hypervisor.get_same_host_hypervisors
get_different_host_hypervisors = _hypervisor.get_different_host_hypervisors
get_server_hypervisor = _hypervisor.get_server_hypervisor
get_servers_hypervisors = _hypervisor.get_servers_hypervisors

find_server_ip_address = _server.find_server_ip_address
HasServerMixin = _server.HasServerMixin
list_server_ip_addresses = _server.list_server_ip_addresses

wait_for_services_up = _service.wait_for_services_up



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os

from tobiko.shell import grep
from tobiko.shell import find
from tobiko.shell import sh


class LogFileDigger(object):

    def __init__(self, filename, **execute_params):
        self.filename = filename
        self.execute_params = execute_params
        self.logfiles = set()
        self.found = set()

    def find_lines(self, pattern, new_lines=False):
        log_files = self.list_log_files()
        try:
            lines = frozenset(
                grep.grep_files(pattern=pattern, files=log_files,
                                **self.execute_params))
        except grep.NoMatchingLinesFound:
            if new_lines:
                return frozenset()
        else:
            lines -= self.found
            self.found.update(lines)
            if new_lines:
                return lines
        return frozenset(self.found)

    def find_new_lines(self, pattern):
        return self.find_lines(pattern=pattern, new_lines=True)

    def list_log_files(self):
        file_path, file_name = os.path.split(self.filename)
        return find.find_files(path=file_path,
                               name=file_name,
                               **self.execute_params)


class JournalLogDigger(LogFileDigger):

    def find_lines(self, pattern, new_lines=False):
        dump_log_cmd = [
            "journalctl", "--unit", self.filename,
            "--since", "5 minutes ago"]
        try:
            lines = frozenset(
                grep.grep_lines(pattern=pattern,
                                command=dump_log_cmd,
                                **self.execute_params))
        except grep.NoMatchingLinesFound:
            if new_lines:
                return frozenset()
        else:
            lines -= self.found
            self.found.update(lines)
            if new_lines:
                return lines
        return frozenset(self.found)


class MultihostLogFileDigger(object):

    def __init__(self, filename, ssh_clients=None,
                 file_digger_class=LogFileDigger,
                 **execute_params):
        self.diggers = collections.OrderedDict()
        self.file_digger_class = file_digger_class
        self.filename = filename
        self.execute_params = execute_params
        if ssh_clients:
            for ssh_client in ssh_clients:
                self.add_host(ssh_client=ssh_client)

    def add_host(self, hostname=None, ssh_client=None):
        hostname = hostname or sh.get_hostname(ssh_client=ssh_client)
        if hostname not in self.diggers:
            self.diggers[hostname] = self.file_digger_class(
                filename=self.filename,
                ssh_client=ssh_client,
                **self.execute_params)

    def find_lines(self, pattern, new_lines=False):
        lines = []
        for hostname, digger in self.diggers.items():
            for line in digger.find_lines(pattern, new_lines=new_lines):
                lines.append((hostname, line))
        return lines

    def find_new_lines(self, pattern):
        return self.find_lines(pattern=pattern, new_lines=True)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko.tests import unit


LOG = log.getLogger(__name__)


class CapureLogTest(unit.TobikoUnitTest):

    def test_capture_log_content(self):
        with tobiko.CaptureLogFixture(test_case_id=self.id()) as capture:
            LOG.warning('< Some debug line>')
        logged_lines = capture.getDetails()['log'].as_text().splitlines()
        expected = f" WARNING {__name__} | < Some debug line>"
        for line in logged_lines:
            if expected in line:
                break
        else:
            lines = '\n'.join(logged_lines)
            self.fail(f"log line not captured: '{expected}' not in \n"
                      f"{lines}")



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko.shell.ping import _ping


LOG = log.getLogger(__name__)


def assert_reachable_hosts(hosts, **params):
    unreachable_hosts = _ping.list_unreachable_hosts(hosts, **params)
    if unreachable_hosts:
        tobiko.fail("Unable to reach host(s): {!r}", unreachable_hosts)


def assert_unreachable_hosts(hosts, **params):
    reachable_hosts = _ping.list_reachable_hosts(hosts, **params)
    if reachable_hosts:
        tobiko.fail("Reached host(s): {!r}", reachable_hosts)



from __future__ import absolute_import

import os
import time
import functools

from oslo_log import log
import pandas
import podman as podmanlib
import docker as dockerlib

import tobiko
from tobiko import podman
from tobiko import docker
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import overcloud
from tobiko.tripleo import topology as tripleo_topology


LOG = log.getLogger(__name__)


def get_container_runtime_module():
    """check what container runtime is running
    and return a handle to it"""
    # TODO THIS LOCKS SSH CLIENT TO CONTROLLER
    ssh_client = topology.list_openstack_nodes(group='controller')[
        0].ssh_client
    if docker.is_docker_running(ssh_client=ssh_client):
        return docker
    else:
        return podman


def get_container_runtime_name():
    return container_runtime_module.__name__.rsplit('.', 1)[1]


if overcloud.has_overcloud():
    container_runtime_module = get_container_runtime_module()
    container_runtime_name = get_container_runtime_name()


@functools.lru_cache()
def list_node_containers(ssh_client):
    """returns a list of containers and their run state"""
    client = get_container_client(ssh_client=ssh_client)
    if container_runtime_module == podman:
        return container_runtime_module.list_podman_containers(client=client)

    elif container_runtime_module == docker:
        return container_runtime_module.list_docker_containers(client=client)


def get_container_client(ssh_client=None):
    """returns a list of containers and their run state"""

    for attempt in tobiko.retry(
            timeout=60.0,
            interval=5.0):
        try:
            if container_runtime_module == podman:
                return container_runtime_module.get_podman_client(
                    ssh_client=ssh_client).connect()
            elif container_runtime_module == docker:
                return container_runtime_module.get_docker_client(
                    ssh_client=ssh_client).connect()
        except dockerlib.errors.DockerException:
            LOG.debug('Unable to connect to docker API')
            attempt.check_limits()
            ssh.reset_default_ssh_port_forward_manager()
    # no successful connection to docker/podman API has been performed
    raise RuntimeError('Unable to connect to container mgmt tool')


def list_containers_df(group=None):
    actual_containers_list = list_containers(group)
    return pandas.DataFrame(
        get_container_states_list(actual_containers_list),
        columns=['container_host', 'container_name', 'container_state'])


def list_containers(group=None):
    """get list of containers in running state
    from specified node group
    returns : a list of overcloud_node's running containers"""

    # moved here from topology
    # reason : Workaround for :
    # AttributeError: module 'tobiko.openstack.topology' has no
    # attribute 'container_runtime'

    containers_list = tobiko.Selection()
    if group:
        openstack_nodes = topology.list_openstack_nodes(group=group)
    else:
        openstack_nodes = topology.list_openstack_nodes(group='overcloud')

    for node in openstack_nodes:
        node_containers_list = list_node_containers(ssh_client=node.ssh_client)
        containers_list.extend(node_containers_list)
    return containers_list


expected_containers_file = '/home/stack/expected_containers_list_df.csv'


def save_containers_state_to_file(expected_containers_list,):
    expected_containers_list_df = pandas.DataFrame(
        get_container_states_list(expected_containers_list),
        columns=['container_host', 'container_name', 'container_state'])
    expected_containers_list_df.to_csv(
        expected_containers_file)
    return expected_containers_file


def assert_containers_running(group, expected_containers, full_name=True):

    """assert that all containers specified in the list are running
    on the specified openstack group(controller or compute etc..)"""

    failures = []

    openstack_nodes = topology.list_openstack_nodes(group=group)
    for node in openstack_nodes:
        node_containers = list_node_containers(ssh_client=node.ssh_client)
        containers_list_df = pandas.DataFrame(
            get_container_states_list(node_containers),
            columns=['container_host', 'container_name', 'container_state'])
        # check that the containers are present
        LOG.info('node: {} containers list : {}'.format(
            node.name, containers_list_df.to_string(index=False)))
        for container in expected_containers:
            # get container attrs dataframe
            if full_name:
                container_attrs = containers_list_df.query(
                    'container_name == "{}"'.format(container))
            else:
                container_attrs = containers_list_df[
                    containers_list_df['container_name'].
                    str.contains(container)]
            # check if the container exists
            LOG.info('checking container: {}'.format(container))
            if container_attrs.empty:
                failures.append(
                    'expected container {} not found on node {} ! : \n\n'.
                    format(container, node.name))
            # if container exists, check it is running
            else:
                # only one running container is expected
                container_running_attrs = container_attrs.query(
                    'container_state=="running"')
                if container_running_attrs.empty:
                    failures.append(
                        'expected container {} is not running on node {} , '
                        'its state is {}! : \n\n'.format(
                            container, node.name,
                            container_attrs.container_state.values.item()))
                elif len(container_running_attrs) > 1:
                    failures.append(
                        'only one running container {} was expected on '
                        'node {}, but got {}! : \n\n'.format(
                            container, node.name,
                            len(container_running_attrs)))

    if failures:
        tobiko.fail('container states mismatched:\n{!s}', '\n'.join(failures))

    else:
        LOG.info('All tripleo common containers are in running state! ')


def assert_all_tripleo_containers_running():
    """check that all common tripleo containers are running
    param: group controller or compute , check containers
    sets in computes or controllers"""

    common_controller_tripleo_containers = ['cinder_api', 'cinder_api_cron',
                                            'cinder_scheduler',
                                            'glance_api', 'heat_api',
                                            'heat_api_cfn',
                                            'heat_api_cron', 'heat_engine',
                                            'horizon', 'iscsid', 'keystone',
                                            'logrotate_crond', 'memcached',
                                            'neutron_api', 'nova_api',
                                            'nova_api_cron', 'nova_conductor',
                                            'nova_metadata', 'nova_scheduler',
                                            'nova_vnc_proxy',
                                            'swift_account_auditor',
                                            'swift_account_reaper',
                                            'swift_account_replicator',
                                            'swift_account_server',
                                            'swift_container_auditor',
                                            'swift_container_replicator',
                                            'swift_container_server',
                                            'swift_container_updater',
                                            'swift_object_auditor',
                                            'swift_object_expirer',
                                            'swift_object_replicator',
                                            'swift_object_server',
                                            'swift_object_updater',
                                            'swift_proxy', 'swift_rsync']

    common_compute_tripleo_containers = ['iscsid', 'logrotate_crond',
                                         'nova_compute', 'nova_libvirt',
                                         'nova_migration_target',
                                         'nova_virtlogd']

    for group, group_containers in [('controller',
                                     common_controller_tripleo_containers),
                                    ('compute',
                                     common_compute_tripleo_containers)]:
        assert_containers_running(group, group_containers)
    # TODO: need to address OSP-version specific containers here.
    # optional ovn containers checks
    assert_ovn_containers_running()


@functools.lru_cache()
def ovn_used_on_overcloud():
    return list_containers_df()['container_name'].\
            str.contains('ovn').any(axis=None)


def assert_ovn_containers_running():
    # specific OVN verifications
    if ovn_used_on_overcloud():
        ovn_controller_containers = ['ovn_controller',
                                     'ovn-dbs-bundle-{}-'.
                                     format(container_runtime_name)]
        ovn_compute_containers = ['ovn_metadata_agent',
                                  'ovn_controller']
        group_containers_list = [('controller', ovn_controller_containers),
                                 ('compute', ovn_compute_containers)]
        if 'networker' in topology.list_openstack_node_groups():
            ovn_networker_containers = ['ovn_controller']
            group_containers_list.append(('networker',
                                          ovn_networker_containers))
        for group, group_containers in group_containers_list:
            assert_containers_running(group, group_containers, full_name=False)
        LOG.info("Networking OVN containers verified in running state")
    else:
        LOG.info("Networking OVN not configured")


def run_container_config_validations():
    """check containers configuration in different scenarios
    """

    # TODO add here any generic configuration validation
    config_checkings = []

    if ovn_used_on_overcloud():
        ovn_config_checkings = \
            [{'node_group': 'controller',
              'container_name': 'neutron_api',
              'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
              'param_validations': [{'section': 'ml2',
                                     'param': 'mechanism_drivers',
                                     'expected_value': 'ovn'},
                                    {'section': 'ml2',
                                     'param': 'type_drivers',
                                     'expected_value': 'geneve'},
                                    {'section': 'ovn',
                                     'param': 'ovn_l3_mode',
                                     'expected_value': 'True'},
                                    {'section': 'ovn',
                                     'param': 'ovn_metadata_enabled',
                                     'expected_value': 'True'}]}]
        config_checkings += ovn_config_checkings
    else:
        ovs_config_checkings = \
            [{'node_group': 'controller',
              'container_name': 'neutron_api',
              'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
              'param_validations': [{'section': 'ml2',
                                     'param': 'mechanism_drivers',
                                     'expected_value': 'openvswitch'}]}]
        config_checkings += ovs_config_checkings

    for config_check in config_checkings:
        for node in topology.list_openstack_nodes(
                group=config_check['node_group']):
            for param_check in config_check['param_validations']:
                obtained_param = sh.execute(
                    f"{container_runtime_name} exec -uroot "
                    f"{config_check['container_name']} crudini "
                    f"--get {config_check['config_file']} "
                    f"{param_check['section']} {param_check['param']}",
                    ssh_client=node.ssh_client, sudo=True).stdout.strip()
                if param_check['expected_value'] not in obtained_param:
                    tobiko.fail(f"Expected {param_check['param']} value: "
                                f"{param_check['expected_value']}\n"
                                f"Obtained {param_check['param']} value: "
                                f"{obtained_param}")
        LOG.info("Configuration verified:\n"
                 f"node group: {config_check['node_group']}\n"
                 f"container: {config_check['container_name']}\n"
                 f"config file: {config_check['config_file']}")


def comparable_container_keys(container, include_container_objects=False):
    """returns the tuple : 'container_host','container_name',
    'container_state, container object if specified'
     """
    if container_runtime_module == podman and include_container_objects:
        return (tripleo_topology.ip_to_hostname(
            container._client._context.hostname),  # pylint: disable=W0212
                container.data['names'], container.data['status'],
                container)
    elif container_runtime_module == podman:
        return (tripleo_topology.ip_to_hostname(
            container._client._context.hostname),  # pylint: disable=W0212
                container.data['names'], container.data['status'])

    elif container_runtime_module == docker and include_container_objects:
        return (container.attrs['Config']['Hostname'],
                container.attrs['Name'].strip('/'),
                container.attrs['State']['Status'],
                container)
    elif container_runtime_module == docker:
        return (container.attrs['Config']['Hostname'],
                container.attrs['Name'].strip('/'),
                container.attrs['State']['Status'])


@functools.lru_cache()
def list_containers_objects_df():
    containers_list = list_containers()
    containers_objects_list_df = pandas.DataFrame(
        get_container_states_list(
            containers_list, include_container_objects=True),
        columns=['container_host', 'container_name',
                 'container_state', 'container_object'])
    return containers_objects_list_df


def get_overcloud_container(container_name=None, container_host=None,
                            partial_container_name=None):
    """gets an container object by name on specified host
    container"""
    con_obj_df = list_containers_objects_df()
    if partial_container_name and container_host:
        con_obj_df = con_obj_df[con_obj_df['container_name'].str.contains(
            partial_container_name)]
        contaniner_obj = con_obj_df.query(
            'container_host == "{container_host}"'.format(
                container_host=container_host))['container_object']
    elif container_host:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'
            ' and container_host == "{container_host}"'.
            format(container_host=container_host,
                   container_name=container_name))['container_object']
    else:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'.
            format(container_name=container_name))['container_object']
    if not contaniner_obj.empty:
        return contaniner_obj.values[0]
    else:
        tobiko.fail('container {} not found!'.format(container_name))


def action_on_container(action,
                        container_name=None, container_host=None,
                        partial_container_name=None):
    """take a container snd preform an action on it
    actions are as defined in : podman/libs/containers.py:14/164"""
    container = get_overcloud_container(
        container_name=container_name,
        container_host=container_host,
        partial_container_name=partial_container_name)
    # we get the specified action as function from podman lib
    if container_runtime_module == podman:
        container_function = getattr(
            podmanlib.libs.containers.Container, '{}'.format(action))
    else:
        container_function = getattr(
            dockerlib.models.containers.Container, '{}'.format(action))
    LOG.info('action_on_container: executing : {} on {}'.format(action,
                                                                container))
    return container_function(container)


def get_container_states_list(containers_list,
                              include_container_objects=False):
    container_states_list = tobiko.Selection()
    container_states_list.extend([comparable_container_keys(
        container, include_container_objects=include_container_objects) for
                                  container in containers_list])
    return container_states_list


pcs_resource_list = ['haproxy', 'galera', 'redis', 'ovn-dbs', 'cinder',
                     'rabbitmq']


def remove_containers_if_pacemaker_resources(comparable_containers_df):
    """remove any containers in
    param: comparable_containers_df that are pacemaker resources
    i.e if they contain tha names of resources defined in
    pcs_resources_list"""

    for row in comparable_containers_df.iterrows():
        for pcs_resource in pcs_resource_list:
            if pcs_resource in str(row):
                LOG.info(f'pcs resource {pcs_resource} has changed state, '
                         f'but that\'s ok since pcs resources can change '
                         f'state/host: {str(row)}')
                # if a pcs resource is found , we drop that row
                comparable_containers_df.drop(row[0], inplace=True)
    return comparable_containers_df


def dataframe_difference(df1, df2, which=None):
    """Find rows which are different between two DataFrames."""
    comparison_df = df1.merge(df2,
                              indicator='same_state',
                              how='outer')
    # return only non identical rows
    if which is None:
        diff_df = comparison_df[comparison_df['same_state'] != 'both']

    else:
        diff_df = comparison_df[comparison_df['same_state'] == which]

    # if the list of diffrent state container are pacemaker resources ignore
    # the error since we are checking pacemaker also.

    remove_containers_if_pacemaker_resources(diff_df)

    return diff_df


def assert_equal_containers_state(expected_containers_list=None,
                                  timeout=120, interval=2,
                                  recreate_expected=False):

    """compare all overcloud container states with using two lists:
    one is current , the other some past list
    first time this method runs it creates a file holding overcloud
    containers' states: /home/stack/expected_containers_list_df.csv'
    second time it creates a current containers states list and
    compares them, they must be identical"""

    # if we have a file or an explicit variable use that , otherwise  create
    # and return
    if recreate_expected or (not expected_containers_list and
                             not os.path.exists(expected_containers_file)):
        save_containers_state_to_file(list_containers())
        return

    elif expected_containers_list:
        expected_containers_list_df = pandas.DataFrame(
            get_container_states_list(expected_containers_list),
            columns=['container_host', 'container_name', 'container_state'])

    elif os.path.exists(expected_containers_file):
        expected_containers_list_df = pandas.read_csv(
            expected_containers_file)

    failures = []
    start = time.time()
    error_info = 'Output explanation: left_only is the original state, ' \
                 'right_only is the new state'

    while time.time() - start < timeout:

        failures = []
        actual_containers_list_df = list_containers_df()

        LOG.info('expected_containers_list_df: {} '.format(
            expected_containers_list_df.to_string(index=False)))
        LOG.info('actual_containers_list_df: {} '.format(
            actual_containers_list_df.to_string(index=False)))

        # execute a `dataframe` diff between the expected and actual containers
        expected_containers_state_changed = \
            dataframe_difference(expected_containers_list_df,
                                 actual_containers_list_df)
        # check for changed state containerstopology
        if not expected_containers_state_changed.empty:
            failures.append('expected containers changed state ! : '
                            '\n\n{}\n{}'.format(
                             expected_containers_state_changed.
                             to_string(index=False), error_info))
            LOG.info('container states mismatched:\n{}\n'.format(failures))
            time.sleep(interval)
            # clear cache to obtain new data
            list_node_containers.cache_clear()
        else:
            LOG.info("assert_equal_containers_state :"
                     " OK, all containers are on the same state")
            return
    if failures:
        tobiko.fail('container states mismatched:\n{!s}', '\n'.join(
            failures))
